""" 
This module was created to handle 
connection in PostgreSQL Databases 
with Python.

"""


from configparser import ConfigParser
from typing import List, Tuple, Union
import psycopg2
import gzip
from drtools.file_manager import path_exist, create_directories_of_path
from drtools.logs import Log
import logging


FetchAll = 'fetch-all'
FetchOne = 'fetch-one'
FetchMany = 'fetch-many'


class Cursor:
    """This class handle psycopg2 connection cursor methods.
    """
    
    def __init__(
        self,
        connection: any,
        fetch_mode: Union[FetchAll, FetchOne, FetchMany]=FetchAll,
        LOGGER: Log=logging
    ) -> None:
        """Handle Psycog2 connection cursor.

        Parameters
        ----------
        connection : _type_
            _description_
        fetch_mode : Union[FetchAll, FetchOne, FetchMany], optional
            _description_, by default FetchAll
        """
        self.LOGGER = LOGGER
        self.LOGGER.info('Initializing cursor...')
        self._cursor = connection.cursor()
        self.fetch_mode = fetch_mode
        self.LOGGER.info('Cursor was successfully initialized!')        
        
    def execute(
        self,
        query: str,
        query_values: Tuple=None,
    ) -> None:
        self.LOGGER.info('Executing query...')
        if query_values is not None:
            self._cursor.execute(query, query_values)
        else:
            self._cursor.execute(query)
        self.LOGGER.info('Query successfully executed...')        
        
    def fetch(
        self, 
        size: int=10
    ) -> List[Tuple]:
        if self.fetch_mode == FetchAll:
            response = self._cursor.fetchall()
        elif self.fetch_mode == FetchOne:
            response = self._cursor.fetchone()
        elif self.fetch_mode == FetchMany:
            assert size is not None, 'If fetch == FetchMany, you need provide "size" value.'
            response = self._cursor.fetchmany(size)
        else:
            raise Exception('Invalid "fetch" mode.')
        return response
    
    def description(self) -> any:
        return self._cursor.description    
    
    def copy(
        self, 
        query: str,
        save_path: str
    ) -> None:        
        outputquery = "COPY ({0}) TO STDOUT WITH CSV HEADER".format(query)
        if path_exist(save_path):
            raise Exception(f'Path {save_path} already exists.')
        create_directories_of_path(save_path)        
        if save_path.endswith('.gz'):
            with gzip.open(save_path,  'w') as f:
                self._cursor.copy_expert(outputquery, f)
        else:
            with open(save_path, 'w') as f:
                self._cursor.copy_expert(outputquery, f)    
    
    def close(self) -> None:
        self._cursor.close()


class ConnectionConfig:
    
    def __init__(
        self,
        sslmode='require',
        connect_timeout: int=3,
        options: str='-c statement_timeout=3000',
        keepalives: int=1,
        keepalives_idle: int=15,
        keepalives_interval: int=10,
        keepalives_count: int=3
    ) -> None:
        """Handle connection configuration.
        
        More information about parameters on https://www.postgresql.org/docs/9.3/libpq-connect.html.

        Parameters
        ----------
        sslmode : str, optional
            The SSL mode that will be 
            applied on connection, by default 'require'
        connect_timeout : int, optional
            Maximum wait for connection, in seconds 
            (write as a decimal integer string). 
            Zero or not specified means wait indefinitely. 
            It is not recommended to use a timeout 
            of less than 2 seconds, by default 3
        options : str, optional
            Adds command-line options to send to the 
            server at run-time. For example, 
            setting this to -c geqo=off sets the 
            session's value of the geqo parameter 
            to off, by default '-c statement_timeout=3000'
        keepalives : Union[On, Off], optional
            Controls whether client-side TCP keepalives 
            are used. The default value is 1, meaning 
            on, but you can change this 
            to 0, meaning off, if keepalives are 
            not wanted. This parameter is ignored 
            for connections made via a 
            Unix-domain socket., by default On
        keepalives_idle : int, optional
            Controls the number of seconds of inactivity 
            after which TCP should send a keepalive 
            message to the server. A value of zero uses the 
            system default. This parameter is ignored 
            for connections made via a Unix-domain 
            socket, or if keepalives are 
            disabled. It is only supported on systems where 
            TCP_KEEPIDLE or an equivalent socket option 
            is available, and on Windows; on other systems, it has 
            no effect, by default 15
        keepalives_interval : int, optional
            Controls the number of seconds after which a TCP 
            keepalive message that is not acknowledged by the server 
            should be retransmitted. A value of zero uses the 
            system default. This parameter is ignored for connections 
            made via a Unix-domain socket, or if keepalives are 
            disabled. It is only supported on systems where 
            TCP_KEEPINTVL or an equivalent socket option is 
            available, and on Windows; on other systems, it has no 
            effect, by default 10
        keepalives_count : int, optional
            Controls the number of TCP keepalives that can 
            be lost before the client's connection to the 
            server is considered dead. A value of zero 
            uses the system default. This parameter is 
            ignored for connections made via a Unix-domain 
            socket, or if keepalives are disabled. It is only 
            supported on systems where TCP_KEEPCNT or an 
            equivalent socket option is available; on other 
            systems, it has no effect, by default 3
        """
        args = locals().copy()
        args = {k: v for k, v in args.items() if k != 'self'}
        for k, v in args.items():
            setattr(self, k, v)


class Database:
    
    def __init__(
        self,
        host: str=None,
        dbname: str=None,
        port: int=None,
        user: str=None,
        password: str=None,
        connection_config: ConnectionConfig=ConnectionConfig(),
        LOGGER: Log=logging
    ) -> None:
        self.LOGGER = LOGGER
        self.host = host
        self.dbname = dbname
        self.port = port
        self.user = user
        self.password = password
        self.connection_config = connection_config
        self._connection = None

    def credentials_config(
        self,
        filename: str, 
        section: str,
    ) -> None:
        # create a parser
        parser = ConfigParser()
        # read config file
        parser.read(filename)

        # get section, default to postgresql
        db = {}
        if parser.has_section(section):
            params = parser.items(section)
            for param in params:
                db[param[0]] = param[1]
        else:
            raise Exception('Section {0} not found in the {1} file'.format(section, filename))
        
        self.host = db.get('host', None)
        self.dbname = db.get('dbname', None)
        self.port = db.get('port', None)
        self.user = db.get('user', None)
        if self.password is None:
            self.password = db.get('password', None)

    def connect(
        self,
    ) -> None:
        """ Connect to the PostgreSQL database server """
        self._connection = None
        try:
            # read connection parameters
            params = {
                'host': self.host,
                'dbname': self.dbname,
                'port': self.port,
                'user': self.user,
                'password': self.password,
            }
            if params['password'] is None:
                raise Exception("You must provice 'password' is required.")            
            
            params = {**params, **self.connection_config.__dict__}
            
            # connect to the PostgreSQL server
            self.LOGGER.info('Connecting to the PostgreSQL database...')
            self._connection = psycopg2.connect(**params)
            self.LOGGER.info('Successful connection!')
            
        except (Exception, psycopg2.DatabaseError) as error:
            self.LOGGER.error(error)                
                
    def execute(
        self, 
        query: str, 
        query_values: Tuple=None,
        fetch_mode: Union[FetchAll, FetchOne, FetchMany]=FetchAll,
        size: int=10,
        save_on: str=None,
    ) -> Tuple[List[str], List[Tuple]]:
        """Execute received query.

        Parameters
        ----------
        query : str
            Text query.
        query_values : Tuple, optional
            Query values, by default None
        fetch_mode : Union[FetchAll, FetchOne, FetchMany], optional
            Fetch mode, by default FetchAll
        size : int, optional
            Fetch size, by default 10
        save_on : str, optional
            Path to save response, by default None

        Returns
        -------
        Tuple[List[str], List[Tuple]]
            Returns the Header and the Body data.
        """
        self._connection = None
        cursor = None
        try:
            self.connect()
            cursor = Cursor(connection=self._connection, fetch_mode=fetch_mode)
            if save_on is None:
                cursor.execute(query, query_values)
                response = cursor.fetch(size)
                colnames = [desc[0] for desc in cursor.description()]
                return colnames, response
            else:
                cursor.copy(query, save_on)
        except (Exception, psycopg2.DatabaseError) as error:
            self.LOGGER.error(error)
        finally:
            if cursor is not None:
                cursor.close()
            self.close()

    def close(self) -> None:
        if self._connection is not None:
                self._connection.close()
                self._connection = None
                self.LOGGER.info('Database connection closed.')


""" 
This module was created to archive useful ways to handle 
with Model transforming.

"""


from pandas import DataFrame
import pandas as pd
import os
import numpy as np
from datetime import datetime
import logging
from drtools.utils import (
    ValueRestrictions, ValueRestrictionsAsJson
)
from drtools.decorators import start_end_log
from drtools.file_manager import (
    create_directories_of_path, rm_file, 
    list_path_of_all_files_inside_directory
)
import drtools.data_science.data_handle as DataHandle
import drtools.utils as Utils
from typing import List, Union, Optional, Dict, TypedDict, Any
from types import FunctionType
from drtools.logs import Log
from drtools.utils import list_ops
from pathlib import Path
from pydbml import PyDBML
from abc import ABC
import ast
from drtools.data_science.data_load import read_dir_as_df, read_as_df
import joblib


Input = 'input'
Output = 'output'
VarChar = 'varchar'
Str = 'str'
Int = 'int'
Float = 'float'
Datetime = 'datetime'
TimeStamp = 'timestamp'
Categorical = 'categorical'
Numerical = 'numerical'


class CategoricalConditions(TypedDict):
    accepted_categories: Dict[str, int]
    accept_others: Optional[bool]
    accept_empty: Optional[bool]
    extra_categories: Optional[Dict[str, int]]


class SimpleFeature(TypedDict):
    name: str
    type: Union[VarChar, Str, Int, Float, Datetime, TimeStamp]    

    
class Feature(SimpleFeature):
    description: Union[Categorical, Numerical]
    conditions: Union[CategoricalConditions, ValueRestrictionsAsJson]
    observation: str


class TypeColumm(TypedDict):
    name: str
    typeraze: FunctionType
    

class Config(TypedDict):
    insert_oneFile: bool
    insert_fileExtension: bool
    typeraze_customTreatment: List[TypeColumm]


def filter_categorical(
    df: DataFrame,
    col: str,
    conditions: dict
) -> DataFrame:
    """Filter categorical data based on conditions.

    Parameters
    ----------
    df : DataFrame
        The DataFrame.
    col : str
        The column from where categories will be filtered.
    conditions : dict
        The conditions that will be applied when filtering categories.

    Returns
    -------
    DataFrame
        The filtered DataFrame.
    """
    
    if len(conditions) == 0:
        return df
    
    df[col] = np.where(pd.isnull(df[col]), df[col], df[col].astype(str))   
            
    accepted_categories = list(conditions.get('accepted_categories').keys())
    
    if conditions.get('accept_others', False):
        join_cat = list(df[col].dropna().unique())
        join_cat = Utils.list_ops(
            join_cat,
            accepted_categories
        )
        df = DataHandle.join_categories(
            df,
            col,
            join_cat,
            'OTHERS'
        )

    if conditions.get('accept_empty', False):
        df[col] = df[col].fillna('EMPTY')
        
    accepted_categories = accepted_categories + list(conditions.get('extra_categories', {}).keys())
    df = df[df[col].isin(accepted_categories)]
    
    return df


def filter_numerical(
    data: Union[DataFrame, np.array],
    cols: Union[List[int], List[str]],
    conditions: dict,
    as_numpy: bool=False
) -> Union[DataFrame, np.array]:
    """Filter numerical data based on conditions.

    Parameters
    ----------
    df : DataFrame
        The DataFrame.
    col : str
        The column from where numerical values will be filtered.
    conditions : dict
        The conditions that will be applied when filtering values.
    as_numpy : bool, Optional
        If True, will apply restrictions on a numpy matrix, 
        If False, will apply restrictions on a DataFrame, 
        by default False. 

    Returns
    -------
    DataFrame
        The filtered DataFrame.
    """
    if not as_numpy:
        if len(conditions) == 0:
            data[cols] = data.loc[:, cols].astype(float)
            return data
        value_restrictions = ValueRestrictions()
        value_restrictions.initialize_from_dict(conditions)
        data[cols] = data.loc[:, cols].astype(float)
        data = value_restrictions.restrict_df(data, cols)
    else:
        if len(conditions) == 0:
            # data[:, col] = data[:, col].astype(float)
            return data
        value_restrictions = ValueRestrictions()
        value_restrictions.initialize_from_dict(conditions)
        data = value_restrictions.restrict_numpy(data, cols)
        # data[:, col] = data[:, col].astype(float)
    return data


String = ['varchar', 'str']
Time = ['timestamp', 'datetime']
Int = ['int']
Float = ['float']
JSONB = ['JSONB', 'JSON']


def typeraze(
    dataframe: DataFrame,
    features: List[SimpleFeature],
    dtypes: List[str]=None,
    ignore_dtypes: List[str]=[],
    custom_treatment: List[TypeColumm]=[],
    utc: Union[bool, None]=None,
    to_numeric: bool=False,
    LOGGER: Log=None,
    **kwargs
) -> DataFrame:
    """Type database columns.

    Parameters
    ----------
    dataframe : DataFrame
        The DataFrame contining the data to be typed.
    features : List[SimpleFeature]
        The features to be typed.
    dtypes : List[str], optional
        Specify Dtypes to be handled, by default None
    ignore_dtypes : List[str], optional
        Ignore Dtypes when typing data, by default []
    custom_treatment : List[TypeColumm], optional
        If some colum needs to have a custom treatment 
        on data, the function to treat the column 
        can be specified here, by default []
    utc : Union[bool, None], optional
        If True, apply pd.to_datetime with utc=True, 
        if False, apply with False, by default None.
    to_numeric : bool, optional
        If True, apply pd.to_numeric before type, 
        if False, do not apply, by default False
    LOGGER : Log, optional
        The Log to verbose, by default None

    Returns
    -------
    DataFrame
        The typed DataFrame

    Raises
    ------
    Exception
        If Dtypes of column is not supported.
    Exception
        When some error occurs when typing some column.
    """
        
    df = dataframe.copy()

    AllDtypes = String + Time + Int + Float + JSONB
    
    if dtypes is None:
        dtypes = AllDtypes
        
    # filter features to process
    string_features_name = []
    time_features_name = []
    int_features_name = []
    float_features_name = []
    JSONB_features_name = []
    real_custom_treatment = []
    
    for feature in features:
        feature_type = feature['type']
        feature_name = feature['name']
        
        is_custom_treatment = False
        for custom_treat_col in custom_treatment:
            if feature_name == custom_treat_col['name']:
                real_custom_treatment.append(custom_treat_col)
                is_custom_treatment = True
                break            
        if is_custom_treatment:
            continue
        
        if feature_type in dtypes \
        and feature_type not in ignore_dtypes:
            
            if feature_type in String:
                string_features_name.append(feature_name)
                
            elif feature_type in Time:
                time_features_name.append(feature_name)
                
            elif feature_type in Int:
                int_features_name.append(feature_name)
                
            elif feature_type in Float:
                float_features_name.append(feature_name)
             
            elif feature_type in JSONB:
                JSONB_features_name.append(feature_name)
                
            elif feature_type not in AllDtypes:
                raise Exception(f'Invalid type {feature_type} on column {feature_name}')
        
    
    LOGGER.debug(f'Typing {len(string_features_name)} cols as String: {string_features_name}...')        
    archive_indexes = df[string_features_name].index
    temp_df = pd.DataFrame(
            np.where(
                pd.isnull(df[string_features_name]), 
                df[string_features_name], 
                df[string_features_name].astype(str)
            ),
            columns=string_features_name
        ).set_index(archive_indexes)
    for col in string_features_name:
        del df[col]
    df = pd.concat([df, temp_df], axis=1)
    LOGGER.debug(f'Typing {len(string_features_name)} cols as String... Done!')
        
    
    LOGGER.debug(f'Typing {len(time_features_name)} cols as Datetime: {time_features_name}...')       
    df[time_features_name] = df[time_features_name].apply(pd.to_datetime, errors='coerce', utc=utc)
    LOGGER.debug(f'Typing {len(time_features_name)} cols as Datetime... Done!')
        
    
    LOGGER.debug(f'Typing {len(int_features_name)} cols as Int64: {int_features_name}...') 
    if to_numeric:
        df[int_features_name] = df[int_features_name].apply(pd.to_numeric, errors='coerce')
    
    try:
        temp_df = df[int_features_name].astype('Int64')
    except:
        df[int_features_name] = df[int_features_name].apply(pd.to_numeric, errors='coerce')
        temp_df = df[int_features_name].astype('Int64')
        
    for col in int_features_name:
        del df[col]
    df = pd.concat([df, temp_df], axis=1)
    del temp_df
    LOGGER.debug(f'Typing {len(int_features_name)} cols as Int64... Done!')
        
        
    LOGGER.debug(f'Typing {len(float_features_name)} cols as Float: {float_features_name}...')
        
    if to_numeric:
        df[float_features_name] = df[float_features_name].apply(pd.to_numeric, errors='coerce')
        
    try:
        temp_df = df[float_features_name].astype(float)
    except:
        df[float_features_name] = df[float_features_name].apply(pd.to_numeric, errors='coerce')
        temp_df = df[float_features_name].astype(float)
        
    for col in float_features_name:
        del df[col]
    df = pd.concat([df, temp_df], axis=1)
    del temp_df
    LOGGER.debug(f'Typing {len(float_features_name)} cols as Float... Done!')
        
        
    LOGGER.debug(f'Typing {len(JSONB_features_name)} cols as JSONB: {JSONB_features_name}...') 
     
    for col in JSONB_features_name:
            
        LOGGER.debug(f'Typing col {col}...') 
                
        is_nan_indexes = df[df[col].isna()].index
        not_nan_indexes = df[df[col].notna()].index
        df_backup = df.copy()
        try:
            df[col] = df[col].astype(str)
            df.loc[df.index.isin(not_nan_indexes), col] \
                = df.loc[df.index.isin(not_nan_indexes), col] \
                    .apply(lambda x: ast.literal_eval(x))
        except:
            df = df_backup
            df.loc[df.index.isin(not_nan_indexes), col] \
                = df.loc[df.index.isin(not_nan_indexes), col] \
                    .apply(lambda x: ast.literal_eval(x) if type(x) == str else x)
            
        df.loc[df.index.isin(is_nan_indexes), col] = np.nan
            
        LOGGER.debug(f'Typing col {col}... Done!') 
        
    LOGGER.debug(f'Typing {len(JSONB_features_name)} cols as JSONB... Done!') 
    
    
    real_custom_treatment_names = [x['name'] for x in real_custom_treatment]
    LOGGER.debug(f'Handling custom treatment on {len(real_custom_treatment)} cols: {real_custom_treatment_names}...') 
    
    for custom_treat_col in real_custom_treatment:
        col = custom_treat_col['name']
        LOGGER.debug(f'Custom treatment on col {col}...') 
        custom_treat_typeraze = custom_treat_col['typeraze']
        df[col] = custom_treat_typeraze(df, col)
        LOGGER.debug(f'Custom treatment on col {col}... Done!') 
        
    LOGGER.debug(f'Handling custom treatment on {len(real_custom_treatment)} cols... Done!') 
    
    return df


class ModelCatalogueSingle(TypedDict):
    id: int
    created_at: datetime
    updated_at: datetime
    name: str
    version: str
    algorithm: str
    algorithm_infrastructure: Any
    description: str
    rules: str
    input_features: List[Feature]
    output_features: List[Feature]


class Model:
    """Class to handle model loading based on definition 
    on definition pattern presented on ModelCatalogue.
    
    Methods
    -------
    - extra_features_name()
    - output_features_name()
    - get_model_name()
    - cols_correct_order()
    - load_model()
    - save_model()
    - train()
    - predict()
    - one_hot_encoding()
    - label_encoding()
    - filter()
    - typeraze()
    """
    
    def __init__(
        self,
        model_catalogue_single: ModelCatalogueSingle,
  		LOGGER: Log=None,
        chained_assignment_log: bool=False
    ) -> None:
        """Init Model instance.

        Parameters
        ----------
        model_catalogue_single : ModelCatalogueSingle
            The model definitions.
        LOGGER : Log, optional
            The LOGGER instance to handle logs 
            , by default None
        chained_assignment_log : bool, optional
            If False, put pandas chained assignment equals None, 
            If True, do not change anything, by default False.
        """
        for k in model_catalogue_single:
            setattr(self, k, model_catalogue_single[k])
        self.LOGGER = logging if LOGGER is None else LOGGER
        if not chained_assignment_log:
            pd.options.mode.chained_assignment = None # default='warn'
            
    # @start_end_log('extra_features_name')
    def extra_features_name(self) -> List[str]:
        """Returns list of model extra columns names.

        Returns
        -------
        List[str]
            Model extra columns names.
        """
        return [feature['name'] for feature in self.extra_features]
    
    # @start_end_log('input_features_name')
    def input_features_name(self) -> List[str]:
        """Returns list of model input features name.

        Returns
        -------
        List[str]
            Model input features name.
        """
        return [feature['name'] for feature in self.input_features]
    
    # @start_end_log('output_features_name')
    def output_features_name(self) -> List[str]:
        """Returns list of model output features name.

        Returns
        -------
        List[str]
            Model output features name.
        """
        return [feature['name'] for feature in self.output_features]
    
    # @start_end_log('get_model_name')        
    def get_model_name(self) -> str:
        """Returns model name.

        Returns
        -------
        str
            Model name combining id, algorithm nickname, model name 
            and model version.
        """
        return f'{self.name}-{self.version}'
    
    # @start_end_log('cols_correct_order')        
    def cols_correct_order(
        self
    ) -> List[str]:
        """Returns list of all cols of model, including 
        extra columns, in correct order.

        Returns
        -------
        List[str]
            Model cols in correct order.
        """
        extra_features_name = self.extra_features_name()
        input_features_name = self.input_features_name()
        output_features_name = self.output_features_name()
        pretty_cols = list_ops(extra_features_name, input_features_name + output_features_name) \
            + input_features_name \
            + output_features_name
        return pretty_cols
    
    @start_end_log('load_model')
    def load_model(
        self,
        model_file_path: str,
        *args,
        **kwargs
    ) -> Any:
        """Load model from path and return model instance

        Parameters
        ----------
        model_file_path : str
            Path of model file.
        args : Tuple, optional
            All args inputs will be passed to load model 
            function, by default ().
        kwargs : Dict, optional
            All args inputs will be passed to load model 
            function, by default {}.

        Returns
        -------
        Any
            The model instance

        Raises
        ------
        Exception
            If model algorithm is invalid
        """
        
        self.LOGGER.info(f'Loading model {self.get_model_name()}...')        
        model = None        
        if self.model_algorithm == 'LightGBM':
            import lightgbm as lgb
            model = lgb.Booster(model_file=model_file_path, *args, **kwargs)
        elif self.model_algorithm == 'NeuralNetworks':
            from tensorflow import keras
            model = keras.models.load_model(model_file_path, *args, **kwargs)
        else:
            raise Exception(f'Algorithm {self.model_algorithm} is invalid.')        
        self.LOGGER.info(f'Loading model {self.get_model_name()}... Done!')        
        return model
    
    @start_end_log('save_model')
    def save_model(
        self,
        model_instance: Any,
        path: str,
        *args,
        **kwargs
    ) -> None:
        """Save model with path.

        Parameters
        ----------
        model_instance : Any
            Instance of desired model to save. 
        path : str
            The path to save model.
        args : Tuple, optional
            All args inputs will be passed to save model 
            function, by default ().
        kwargs : Dict, optional
            All args inputs will be passed to save model 
            function, by default {}.

        Returns
        -------
        None
            None is returned.

        Raises
        ------
        Exception
            If model algorithm is invalid
        """
        self.LOGGER.info(f'Saving model {self.get_model_name()}...')        
        # save_path = f'{project_root_path}/models/{self.get_model_name()}/model/{self.get_model_name()}'        
        if self.model_algorithm == 'LightGBM':
            create_directories_of_path(path)
            model_instance.save_model(filename=path, *args, **kwargs)
        elif self.model_algorithm == 'NeuralNetworks':
            create_directories_of_path(path)
            model_instance.save(path, *args, **kwargs)
        else:
            raise Exception(f'Algorithm {self.model_algorithm} is invalid.')        
        self.LOGGER.info(f'Saving model {self.get_model_name()}... Done!')
    
    @start_end_log('train')
    def train(
        self,
        model_instance: Any,
        *args,
        **kwargs
    ) -> Any:
        """Train model.

        Parameters
        ----------
        model_instance : Any
            Instance of desired model to train.
        args : Tuple, optional
            All args inputs will be passed to train 
            function, by default ().
        kwargs : Dict, optional
            All kwargs inputs will be passed to train 
            function, by default {}.

        Returns
        -------
        Any
            Returns different data for each algorithm. 

        Raises
        ------
        Exception
            If model algorithm is invalid
        """
        self.LOGGER.info(f'Training model {self.get_model_name()}...')                
        if self.model_algorithm == 'LightGBM':
            import lightgbm as lgb
            model_instance = lgb.train(*args, **kwargs)
            self.LOGGER.info(f'Training model {self.get_model_name()}... Done!')            
            return model_instance
        elif self.model_algorithm == 'NeuralNetworks':
            history = model_instance.fit(*args, **kwargs)
            self.LOGGER.info(f'Training model {self.get_model_name()}... Done!')            
            return model_instance, history
        else:
            raise Exception(f'Algorithm {self.model_algorithm} is invalid.')
    
    @start_end_log('predict')
    def predict(
        self,
        model_file_path: str,
        X: Any,
        *args,
        **kwargs
    ) -> Any:    
        """Predict data.

        Parameters
        ----------
        model_file_path : str
            Path of model file.
        X : Any
            X data to predict.
        args : Tuple, optional
            All args inputs will be passed to predict 
            function, by default ().
        kwargs : Dict, optional
            All kwargs inputs will be passed to predict 
            function, by default {}.

        Returns
        -------
        Any
            Returns different data for each algorithm. 

        Raises
        ------
        Exception
            If model algorithm is invalid
        """    
        self.LOGGER.info(f'Predicting data for model {self.get_model_name()}...')  
        model_instance = self.load_model(model_file_path)
        if self.model_algorithm == 'LightGBM':
            y_pred = model_instance.predict(X, *args, **kwargs)
        elif self.model_algorithm == 'NeuralNetworks':
            y_pred = model_instance.predict(X, *args, **kwargs)
            y_pred = y_pred.reshape(1, -1)[0]
        else:
            raise Exception(f'Algorithm {self.model_algorithm} is invalid.')        
        self.LOGGER.info(f'Predicting data for model {self.get_model_name()}... Done!')        
        return y_pred
    
    @start_end_log('one_hot_encoding')
    def one_hot_encoding(
        self,
        dataframe: DataFrame,
        encode_cols: List[str]
    ) -> DataFrame:
        """One hot encode variables, drop original column that 
        generate encoded and drop dummy cols that is not present 
        on the input features.
        
        Parameters
        ----------
        dataframe : DataFrame
            DataFrame containing data to encode.
        encode_cols : List[str]
            List with name of columns to one hot encode.
            
        Returns
        -------
        DataFrame
            The DataFrame containing encoded columns.
        """
        df = dataframe.copy()        
        for col in encode_cols:
            curr_features = [
                feature for feature in self.input_features
                if feature.get('observation', None) == col
            ]
            dummies = pd.get_dummies(df[col], prefix=col)
            drop_cols = list_ops(dummies.columns, self.input_features_name())
            df = pd.concat([df, dummies], axis=1)
            drop_self_col = list_ops([col], self.extra_features_name())            
            df = df.drop(drop_cols + drop_self_col, axis=1)            
            # insert feature that not has on received dataframe
            for curr_feature in curr_features:
                if curr_feature['name'] not in df.columns:
                    df[curr_feature['name']] = 0
        return df
    
    @start_end_log('label_encoding')
    def label_encoding(
        self,
        dataframe: DataFrame,
        astype_category: bool=False
    ) -> DataFrame:
        """Label encode variables.
        
        Parameters
        ----------
        dataframe : DataFrame
            DataFrame containing data to encode.
        astype_category : bool, optional
            If True, in set categorical columns to type "category", 
            If False, will encode values with integers, by default False
            
        Returns
        -------
        DataFrame
            The DataFrame containing encoded columns.
        """
        df = dataframe.copy()
        categorical = {
            feature['name']: feature['conditions']
            for feature in self.input_features 
            if feature.get('description', None) == Categorical
        }        
        encode = {
            col: {
                **conditions.get('accepted_categories'),
                **conditions.get('extra_categories', {})
            }
            for col, conditions in categorical.items()
        }        
        temp_encode = {
            alias: {
                **conditions.get('accepted_categories'),
                **conditions.get('extra_categories', {})
            }
            for conditions in categorical.values()
            for alias in conditions.get('aliases', [])
        }        
        encode = { **encode, **temp_encode }            
        for col in df.columns:
            if col in encode:
                if astype_category:
                    df[col] = df[col].astype('category')
                else:
                    df[col] = df[col].astype(str)
                    df[col] = df[col].replace(encode[col])        
        return df
    
    @start_end_log('filter')
    def filter(
        self,
        dataframe: DataFrame,
        filter_only: List[str]=None,
        ignore_features: List[str]=[],
        ignore_categorical: bool=False,
        ignore_numerical: bool=False,
        custom: List[Feature]=None,
        verbosity: bool=True,
        as_numpy: bool=False,
    ) -> DataFrame:
        """Filter data to acceptable data to model.

        Parameters
        ----------
        dataframe : DataFrame
            The DataFrame containing data to be filtered.
        filter_only : List[str], optional
            Filter only features that has name listed on this variable 
            , by default None.
        ignore_features : List[str], optional
            Ignore all features that has name listed on this variable 
            , by default [].
        ignore_categorical : bool, optional
            If True, ignore all Categorical features, by default False.
        ignore_numerical : bool, optional
            If True, ignore all Numerical features, by default False.
        custom : List[Feature], optional
            Custom list of features to filter. If provide, 
            will filter only the features listed on this variable, by default None.
        verbosity : bool, optional
            If True, verbose function statements, 
            if False, do not verbose, by default True.
        as_numpy : bool, optional
            If True, will treat numerical data as numpy arrays, 
            this option can decrease the computation time a lot 
            , by default False.

        Returns
        -------
        DataFrame
            The filtered DataFrame.

        Raises
        ------
        Exception
            If "description" fielf of Feature is invalid.
        """
        
        df = dataframe.copy()
        
        try:
            self.LOGGER.set_verbosity(verbosity)
        except:
            pass
        
        all_features = self.input_features + self.output_features
        
        filter_features = all_features
        if custom is not None:
            filter_features = custom
            
        categorical_features = []
        unique_numerical_conditions = []
        
        for feature in filter_features:
            
            if feature["name"] in ignore_features:
                self.LOGGER.debug(f'Ignore feature {feature["name"]}. Skipping...')
                continue
            
            if filter_only is not None:
                if feature["name"] not in filter_only:
                    self.LOGGER.debug(f'Feature {feature["name"]} not on Filter Only. Skipping...')
                    continue
            
            if feature["description"] == Numerical:
                
                if ignore_numerical:
                    self.LOGGER.debug(f'Ignore Numerical feature {feature["name"]}. Skipping...')
                    continue
                
                already_inserted = False
                for idx, row in enumerate(unique_numerical_conditions):
                    conditions = row[0]
                    if feature["conditions"] == conditions:
                        unique_numerical_conditions[idx][1].append(feature)
                        already_inserted = True
                        break
                    
                if not already_inserted:
                    unique_numerical_conditions.append(
                        [feature["conditions"], [feature]]
                    )
                    
                
            elif feature["description"] == Categorical:
                
                if ignore_categorical:
                    self.LOGGER.debug(f'Ignore Categorical feature {feature["name"]}. Skipping...')
                    continue
            
                categorical_features.append(feature)
                
            else:
                raise Exception(f'Feature {feature} has an invalid "description" field.')
            
        self.LOGGER.debug('Transforming and filtering Numerical variables...')
        
        numerical_unique_conditions_len = len(unique_numerical_conditions)        
        df_columns = df.columns
            
        self.LOGGER.debug('Generating numerical data...')
        if as_numpy:
            numerical_data = df.values
            numerical_data[pd.isnull(numerical_data)] = np.nan
        else:
            numerical_data = df
        self.LOGGER.debug('Generating numerical data... Done!')
            
        for idx, row in enumerate(unique_numerical_conditions):
            
            conditions = row[0]
            features = row[1]
            curr_features_name = [x['name'] for x in features]
            count = idx + 1
                
            self.LOGGER.debug('')
            self.LOGGER.debug(f'({count:,}/{numerical_unique_conditions_len:,}) Filtering {len(curr_features_name):,} numerical features...')
            self.LOGGER.debug(f'Features: {curr_features_name}...')
            self.LOGGER.debug(f'Conditions: {conditions}')
            
            from_shape = numerical_data.shape
                            
            if as_numpy:
                numerical_data = filter_numerical(
                    numerical_data, 
                    [df_columns.get_loc(name) for name in curr_features_name], 
                    conditions,
                    as_numpy=True
                )
            else:
                numerical_data = filter_numerical(
                    numerical_data, 
                    curr_features_name, 
                    conditions
                )
                
            to_shape = numerical_data.shape
            self.LOGGER.debug(f'Filter: from shape ({from_shape[0]:,}, {from_shape[1]:,}) to shape ({to_shape[0]:,}, {to_shape[1]:,})')
            self.LOGGER.debug(f'({count:,}/{numerical_unique_conditions_len:,}) Filtering {len(curr_features_name):,} numerical features... Done!')
            
            
        if as_numpy:
            self.LOGGER.debug('Generating DataFrame after filter numerical...')
            df = pd.DataFrame(
                numerical_data,
                columns=df_columns
            )
            self.LOGGER.debug('Generating DataFrame after filter numerical... Done!')
        else:
            df = numerical_data
            
        self.LOGGER.debug('Transforming and filtering Numerical variables... Done!')
        self.LOGGER.debug('')
        self.LOGGER.debug('Transforming and filtering Categorical variables...')
        
        categorical_features_len = len(categorical_features)
        
        for idx, feature in enumerate(categorical_features):
            
            count = idx + 1
                
            self.LOGGER.debug('')
            self.LOGGER.debug(f'({count:,}/{categorical_features_len:,}) Filtering {feature["name"]}...')
            self.LOGGER.debug(f'Conditions: {feature["conditions"]}')            
            
            from_shape = df.shape
            df = filter_categorical(df, feature["name"], feature["conditions"])                
            to_shape = df.shape
                
            self.LOGGER.debug(f'Col <{feature["name"]}> filter from shape ({from_shape[0]:,}, {from_shape[1]:,}) to shape ({to_shape[0]:,}, {to_shape[1]:,})')                    
            self.LOGGER.debug(f'({count:,}/{categorical_features_len:,}) Filtering {feature["name"]}... Done!')
            
        self.LOGGER.debug('Transforming and filtering Categorical variables... Done!')
            
        try:
            self.LOGGER.reset_verbosity()
        except:
            pass
            
        return df
    
    @start_end_log('typeraze')
    def typeraze(
        self,
        dataframe: DataFrame,
        as_model_input: bool=False,
        type_only: List[str]=None,
        ignore_features: List[str]=[],
        custom: List[Feature]=None,
        **kwargs
    ) -> DataFrame:
        """Type model variables.

        Parameters
        ----------
        dataframe : DataFrame
            The DataFrame containing the data to be typed.
        as_model_input : bool, optional
            If True, when Feature has "type" equals "int", treat as "float", 
            if False, when Feature has "type" equals "int", treat as "int", 
            by default False.
        type_only : List[str], optional
            Type only features that has name listed on this variable 
            , by default None.
        ignore_features : List[str], optional
            Ignore all features that has name listed on this variable 
            , by default [].
        custom : List[Feature], optional
            Custom list of features to type. If provide, 
            will type only the features listed on this variable, by default None.
        kwargs : Dict, optional
            Arguments that will be passed to read_dir_as_df() 
            function and to typeraze() function 
            , by default {}.

        Returns
        -------
        DataFrame
            The typed DataFrame.
        """
        all_features_name = self.cols_correct_order()
        
        input_features_name = self.input_features_name()
        output_features_name = self.output_features_name()
        extra_features_name = self.extra_features_name()
        
        all_features = []
        for feature_name in all_features_name:
            if feature_name in input_features_name:
                for feature in self.input_features:
                    if feature_name == feature['name']:
                        all_features.append(feature)
                        break
            elif feature_name in output_features_name:
                for feature in self.output_features:
                    if feature_name == feature['name']:
                        all_features.append(feature)
                        break
            elif feature_name in extra_features_name:
                for feature in self.extra_features:
                    if feature_name == feature['name']:
                        all_features.append(feature)
                        break
            else:
                raise Exception(f'Feature "{feature_name} not present on input features, output features and extra columns"')
        
        if type_only is None:
            type_only = all_features_name
        
        features = [
            {
                'name': col['name'], 
                'type': col['type'] if not as_model_input \
                    else 'float' if col['type'] == 'int' \
                        else col['type'] 
            } 
            for col in all_features
            if col['name'] not in ignore_features \
            and col['name'] in type_only
        ]
        
        if custom is not None:
            features = [
                {'name': feature['name'], 'type': feature['type']}
                for feature in custom
            ]
        df = dataframe.copy()
        df = typeraze(dataframe=df, features=features, LOGGER=self.LOGGER, **kwargs)
        return df


class Database(ABC):
    """Abstract class to handle database loading based on default 
    infrastructure.
    
    Methods
    -------
    - get_col_names()
    - typeraze()
    - load_db_as_df()
    - _save_data()
    - insert()
    - delete_by_id()
    - update_by_id()
    """
    
    CONFIG: Config = Config({
        'insert_oneFile': False,
        'insert_fileExtension': 'parquet',
        'typeraze_customTreatment': []
    })
    
    def __init__(
        self,
        database_dbml_path: str,
        database_root_path: str,
  		LOGGER: Log=None,
    ) -> None:
        """Init Database instance.

        Parameters
        ----------
        database_dbml_path : str
            Path to .dbml database documentation.
        database_root_path : str
            Path to root of database.
        LOGGER : Log, optional
            The LOGGER instance to handle logs 
            , by default None
        """
        assert hasattr(self, 'SCHEMA')
        assert hasattr(self, 'TABLE')  
        self.database_dbml_path = database_dbml_path
        self.database_root_path = database_root_path
        parsed = PyDBML(Path(database_dbml_path))
        try:
            self.table = parsed[f'{self.SCHEMA}.{self.TABLE}']
            self.col_names = [col.name for col in self.table.columns]
        except:
            self.enum = [
                enum 
                for enum in parsed.enums 
                if enum.name == self.TABLE \
                and enum.schema == self.SCHEMA
            ][0]
            self.col_names = []
            for col in self.enum.items:
                self.col_names.append(col.name)
                setattr(self, col.name, col)
            
            
        self.LOGGER = logging if LOGGER is None else LOGGER
        
    def get_col_names(self) -> List[str]:
        """Get database column names.

        Returns
        -------
        List[str]
            List containing database column names.
        """
        return self.col_names
    
    def get_manual_col_names(self) -> List[str]:
        """Get database manual column names.

        Returns
        -------
        List[str]
            List containing database column names.
        """
        resp = list_ops(
            self.col_names,
            ['id', 'created_at', 'updated_at']
        )
        return resp      
    
    def db_name(self) -> str:
        """Get name of database

        Returns
        -------
        str
            Name of database
        """
        return f'{self.SCHEMA}.{self.TABLE}'
    
    @start_end_log('typeraze')
    def typeraze(
        self,
        dataframe: DataFrame,
        dtypes: List[str]=None,
        ignore_dtypes: List[str]=[],
        verbosity: bool=False,
        **kwargs
    ) -> DataFrame:
        """Type database columns.

        Parameters
        ----------
        dataframe : DataFrame
            The DataFrame contining the data to be typed.
        dtypes : List[str], optional
            Specify Dtypes to be handled, by default None
        ignore_dtypes : List[str], optional
            Ignore Dtypes when typing data, by default []
        verbosity : bool, optional
            If True, verbose functions statements, 
            if False, do not verbose, by default True

        Returns
        -------
        DataFrame
            The typed DataFrame

        Raises
        ------
        Exception
            If Dtypes of column is not supported.
        Exception
            When some error occurs when typing some column.
        """
            
        df = dataframe.copy()
        useful_cols = list_ops(
            df.columns,
            self.col_names,
            ops='intersection'
        )
        df = df[useful_cols]
        
        features = [
            {'name': col.name, 'type': col.type} 
            for col in self.table.columns 
            if col.name in useful_cols
        ]
        
        custom_treatment = self.CONFIG.get('typeraze_customTreatment', [])
        
        try:
            self.LOGGER.set_verbosity(verbosity)
        except:
            pass
        
        df = typeraze(
            dataframe=df,
            features=features,
            dtypes=dtypes,
            ignore_dtypes=ignore_dtypes,
            custom_treatment=custom_treatment,
            LOGGER=self.LOGGER,
            **kwargs
        )
        
        try:
            self.LOGGER.reset_verbosity()
        except:
            pass
        
        return df
        
    # @start_end_log('load_db_as_df')
    def load_db_as_df(
        self, 
        typeraze: bool=False,
        verbosity: bool=False,
        **kwargs
    ) -> DataFrame:
        """Load local database data as DataFrame

        Parameters
        ----------
        typeraze : bool, optional
            If True, the data will be typed after 
            loaded. If False, the data will not be 
            typed after load, by default False
        kwargs : Dict, optional
            Arguments that will be passed to read_dir_as_df() 
            function and to typeraze() function 
            , by default {}.

        Returns
        -------
        DataFrame
            The local database data as DataFrame
        """
        self.LOGGER.info(f'Loading {self.db_name()}...')
        
        df = read_dir_as_df(
            Utils.join_path(
                self.database_root_path, self.SCHEMA, self.TABLE
            ),
            **kwargs
        )
        
        if df is None:
            real_cols = kwargs.get('usecols', None)
            if real_cols is None:
                real_cols = self.get_col_names()
            df = pd.DataFrame([], columns=real_cols)
          
        if typeraze:              
            df = self.typeraze(
                df,
                verbosity=verbosity,
                **kwargs
            )
        
        cols_correct_order = list_ops(self.get_col_names(), df.columns, ops='intersection')            
        df = df.loc[:, cols_correct_order]
            
        self.LOGGER.info(f'Loading {self.db_name()}... Done!')
            
        return df
    
    @start_end_log('_save_data')
    def _save_data(
        self,
        dataframe: DataFrame,
        filename: str,
        typeraze: bool=True,
        overwrite_if_exists: bool=False
    ) -> None:
        """Save data on local Database infrastructure handling types 
        of columns and filtering columns to fit with database Schema.

        Parameters
        ----------
        dataframe : DataFrame
            The DataFrame to be saved.
        filename : str
            The desired filename to save the data.
        typeraze : bool, optional
            If the data needs to be typed before 
            save, by default True
        overwrite_if_exists : bool, optional
            If True, overwrite file with same name 
            if exists. If False, return Exception 
            if exist some file on local Database with 
            same filename, by default False

        Raises
        ------
        Exception
            If save path already exists.
        Exception
            If extension of file is not allowed.
        """
        
        df = dataframe.copy()
        df = df.loc[:, self.get_col_names()]
        if typeraze:
            df = self.typeraze(df, verbosity=False)
        df = df.loc[:, self.get_col_names()]
        
        save_path = Utils.join_path(
            self.database_root_path, self.SCHEMA, self.TABLE, filename
        )
        if not overwrite_if_exists:
            if os.path.exists(save_path):
                raise Exception(f'Path {save_path} already exists.')
        
        create_directories_of_path(save_path)
        
        if '.csv' in filename:
            df.to_csv(save_path, index=False)
            
        elif '.parquet' in filename:
            
            for col in df.columns:                
                real_col = [col1 for col1 in self.table.columns if col == col1.name][0]
                if real_col.type in JSONB:
                    df[col] = df[col].astype(str)
                    
            df.to_parquet(save_path, index=False)
            
        elif '.json' in filename:
            df.to_json(
                save_path,                
                orient='records', 
                date_format='iso',
                indent=4
            )
        else:
            raise Exception(f'Extension on {filename} not allow.')
    
    # @start_end_log('insert')
    def insert(
        self,
        dataframe: DataFrame
    ) -> List[int]:
        """Insert data on db. Handle data before insert. Typeraze and 
        get only columns that must be exist on db.

        Parameters
        ----------
        dataframe : DataFrame
            The data in DataFrame format to be inserted

        Returns
        -------
        List[int]
            The inserted ids.
        """
        
        self.LOGGER.info(f'Inserting data on {self.db_name()}...')
        
        df = dataframe.copy()
        
        if df.shape[0] == 0:
            return []
        
        already_on_db = self.load_db_as_df(
            usecols=['id']
        )
        
        if already_on_db.shape[0] == 0:
            max_id = 0
        else:
            max_id = already_on_db.id.max()
        
        df['id'] = max_id + np.arange(len(df)) + 1
        inserted_ids = df['id'].values
        df['created_at'] = datetime.now().isoformat()
        df['updated_at'] = datetime.now().isoformat()
        
        # insert data
        default_extension = 'parquet'
        extension = self.CONFIG.get('insert_fileExtension', None)
        extension = default_extension if extension is None else extension
        filename = f'{self.SCHEMA}_{self.TABLE}-{datetime.now().isoformat()}.{extension}'
        
        if self.CONFIG.get('insert_oneFile', False):
            
            already_on_db = self.load_db_as_df()
            df = pd.concat(
                [already_on_db, df], 
                axis=0,
                ignore_index=True
            )
            self._save_data(df, filename)
            
            data_path = Utils.join_path(
                self.database_root_path, self.SCHEMA, self.TABLE
            )
            paths = list_path_of_all_files_inside_directory(data_path)
            for path in paths:
                if path.endswith(filename):
                    continue
                rm_file(
                    path,
                    ignore_if_path_not_exists=True
                )
        else:
            self._save_data(df, filename)
        
        self.LOGGER.info(f'Inserting data on {self.db_name()}... Done!')
        
        return sorted(inserted_ids)
    
    def insert_if_not_exists(
        self,
        dataframe: DataFrame
    ) -> List[int]:
        """Insert data on db only if not exists. Handle data before insert. Typeraze and 
        get only columns that must be exist on db.

        Parameters
        ----------
        dataframe : DataFrame
            The data in DataFrame format to be inserted

        Returns
        -------
        List[int]
            The inserted ids.
        """
        
        
        df = dataframe.copy()
        
        if df.shape[0] == 0:
            return []        
        
        self.LOGGER.info(f'Filtering already inserted...')
        original_shape = df.shape
        necessary_cols = list_ops(
            self.get_col_names(),
            ['id', 'created_at', 'updated_at']
        )
        df = df.loc[:, necessary_cols]
        df['md5_of_row'] = df.apply(lambda x: joblib.hash(x.values), axis=1)
        intersection_data = self.load_db_as_df(
            chunksize=400 * 10**3,
            process_chunk=lambda chunk_df: chunk_df[
                chunk_df.apply(
                    lambda x: joblib.hash(x.values), axis=1
                ).isin(df.md5_of_row.values)
            ]
        )
        intersection_data['md5_of_row'] = intersection_data.apply(
            lambda x: joblib.hash(x.values), axis=1
        )
        df = df[~df.md5_of_row.isin(intersection_data.md5_of_row.values)]
        df = df.drop('md5_of_row', axis=1)
        del intersection_data
        final_shape = df.shape
        self.LOGGER.info(f'{(original_shape[0] - final_shape[0]):,} rows was already inserted.')
        self.LOGGER.info(f'{final_shape[0]:,} rows to insert.')
        self.LOGGER.info(f'Filtering already inserted... Done!')
        
        # return df
        
        inserted_ids = self.insert(
            dataframe=df
        )
        
        return inserted_ids
        
    @start_end_log('delete_by_id')
    def delete_by_id(
        self,
        ids: List[int]
    ) -> List[int]:
        """Delete data from local Database by ids.

        Parameters
        ----------
        ids : List[int]
            The list of ids to be deleted.

        Returns
        -------
        List[int]
            The list of ids that was deleted.
        """
        
        data_path = Utils.join_path(
            self.database_root_path, self.SCHEMA, self.TABLE
        )
        paths = list_path_of_all_files_inside_directory(data_path)
        for path in paths:
            df = read_as_df(path)
            old_shape = df.shape
            df = df[~df.id.isin(ids)]
            new_shape = df.shape
            filename = os.path.basename(path)
            
            if old_shape[0] == new_shape[0]:
                continue
            
            self._save_data(
                df, 
                filename, 
                typeraze=False,
                overwrite_if_exists=True
            )
        
        return sorted(list(ids))
    
    @start_end_log('update_by_id')
    def update_by_id(
        self,
        dataframe: DataFrame
    ) -> List[int]:
        """Update data from local Database with id and data 
        from received DataFrame.

        Parameters
        ----------
        dataframe : DataFrame
            The data in DataFrame format to be inserted

        Returns
        -------
        List[int]
            The updated ids.
        """
        
        df = dataframe.copy()
        
        if df.shape[0] == 0:
            return []
        
        intersect_cols = list_ops(df.columns, self.get_col_names(), ops='intersection')
        intersect_cols = list_ops(intersect_cols, ['created_at', 'updated_at'])
        df = df.loc[:, intersect_cols]
        intersect_cols = intersect_cols + ['updated_at']
        df['updated_at'] = datetime.now().isoformat()
        
        if 'id' not in intersect_cols:
            raise Exception('Column id must be on DataFrame.')
        
        if df.id.values.shape[0] != df.id.unique().shape[0]:
            seen = set()
            dupes = [x for x in df.id.values if x in seen or seen.add(x)]
            dupes_df = df[df.id.isin(dupes)]
            self.LOGGER.info(dupes_df)
            raise Exception(f'There are duplicated ids on received data. Duplicate ids: {dupes}')
        
        data_path = Utils.join_path(
            self.database_root_path, self.SCHEMA, self.TABLE
        )
        paths = list_path_of_all_files_inside_directory(data_path)
        
        updated_ids = []
        
        for path in paths:
            database = read_as_df(path)
            
            ids_intersection = list_ops(
                database.id.values,
                df.id.values, 
                ops='intersection'
            )
            
            complementary_df = database.loc[~database.id.isin(ids_intersection)]
            database = database.loc[database.id.isin(ids_intersection)]
            
            curr_updated_ids = database.id.values.copy()
            
            drop_cols = list_ops(self.get_col_names(), df.columns, ops='intersection')
            drop_cols = list_ops(drop_cols, ['id'])
            
            database = database.drop(drop_cols, axis=1)
            database = database.merge(df, on=['id'], how='inner')

            database = database.loc[:, self.get_col_names()]
            database = pd.concat([
                complementary_df[self.get_col_names()], 
                database[self.get_col_names()]
            ], axis=0)
            database = database.sort_values('id').reset_index(drop=True)
 
            updated_ids = updated_ids + list(curr_updated_ids)
            
            filename = os.path.basename(path)
            
            self._save_data(
                database, 
                filename, 
                typeraze=True,
                overwrite_if_exists=True
            )
            
        not_updated_ids = list_ops(
            df.id.values,
            updated_ids
        )
        
        if len(not_updated_ids) > 0:
            not_updated_ids = sorted(not_updated_ids)
            self.LOGGER.debug(f'Not updated ids: {not_updated_ids}')
            
                    
        return sorted(updated_ids)
    
    def get_by_id(
        self,
        id: int,
        as_dataframe: bool=True
    ):
        df = self.load_db_as_df(
            chunksize=400 * 10**3,
            process_chunk=lambda chunk: chunk[chunk.id == id]
        )
        if as_dataframe:
            return df
        else:
            single = df.to_dict(orient='records')
            return single[0]
    

class ModelCatalogueWhen(TypedDict):
    id: int
    input_features: List[str]
    output_features: List[str]


First = 'first'
All = 'all'

    
class ModelCatalogue(Database):
    
    SCHEMA = 'model'
    
    TABLE = 'catalogue'
    
    @start_end_log('get_model')
    def get_model(
        self,
        when: ModelCatalogueWhen,
        returns: Union[First, All]=First
    ) -> Union[ModelCatalogueSingle, List[ModelCatalogueSingle]]:
        """Get model as dict from model.catalogue database by id.

        Parameters
        ----------
        when : ModelCatalogueWhen
            Conditions that model must satisfy.
        returns : Union[First, All], optional
            If "first", will returns only the first match of criterias, 
            If "all", will returns all models that match the criterias, 
            by default First

        Returns
        -------
        Union[Dict, List[Dict]]
            Dict representing the model or List of model dicts.

        Raises
        ------
        Exception
            If returns option is invalid.
        """
        df = self.load_db_as_df(typeraze=True)
        if 'id' in when:
            model_definition = df[df.id == when['id']]
        model_results = model_definition.to_dict(orient='records')
        if returns == First:
            return model_results[0]
        elif returns == All:
            return model_results
        else:
            raise Exception(f'Given value {returns} for "returns" variable is invalid.')



""" 
This module was created to archive useful 
functions which can be used 
in drtools.data_science.database 
package.

"""


from typing import List


def get_queries_from_sql_file(
    filepath: str,
    split: str=';'
) -> List[str]:
    """Get SQL queries from a SQL file.

    Parameters
    ----------
    filepath : str
        Path to SQL file.

    Returns
    -------
    List[str]
        List of queries found in SQL file.
    """
    # Open and read the file as a single buffer
    fd = open(filepath, 'r')
    sql_file = fd.read()
    fd.close()
    if split is not None:
        # all SQL commands (split on ';')
        sql_commands = sql_file.split(';')
    else:
        sql_commands = sql_file.split(';')
    return sql_commands


""" 
This module was created to manage files and directories
of system.

"""

import os
from typing import List
from drtools.utils import (
    get_os_name,
    join_path
)
import json
import gzip


def path_exist(
    path: str
) -> bool: 
    """Verify if path exists.

    Parameters
    ----------
    path : str
        Path to be verified

    Returns
    -------
    bool
        If True, path exists, if False, path does
        not exists.
    """
    return os.path.exists(path)


def split_path(
    path: str
) -> List[str]:
    """Handle path split by of os.

    Parameters
    ----------
    path : str
        Path to be splited

    Returns
    -------
    List[str]
        List with pieces of path.
    """
    if get_os_name() == "Windows": return path.split("\\")
    elif get_os_name() == "Linux": return path.split("/")
    elif get_os_name() == "Darwin": return path.split("\\")


def move_file(
    current_path: str,
    destination_path: str,
) -> None:
    """Move file handling when file already exists.

    Parameters
    ----------
    current_path : str
        Current path location
    destination_path : str
        Destination path

    Raises
    ------
    Exception
        If current path does not exists
    Exception
        If destination path already exists
    """
    
    if not path_exist(current_path):
        raise Exception('Current path does not exists.')
    if path_exist(destination_path):
        raise Exception('Destination path already exists.')   
    
    create_directories_of_path(destination_path)
    os.rename(current_path, destination_path)


def get_files_path(
    parent_path: str, 
    content_arr: List[str]=[],
    ignore_files: List[str]=[],
    ignore_folders: List[str]=[],
    ignore_if_contain_some_of_the_words: List[str]=[]
) -> List:
    """Return array containing all abs path of files inside directory.

    Parameters
    ----------
    parent_path : str
        Directory to search files.
    content_arr : List[str], optional
        Array of content of parent path directory, by default []
    ignore_files : List[str], optional
        Ignore files if name equal some of this list, by default [] 
    ignore_folders : List[str], optional
        Ignore folders if name equal some of this list, by default []
    ignore_if_contain_some_of_the_words : List[str], optional
        Ignore folders and files if name cotain some of 
        the names presented on this list, by default []

    Returns
    -------
    List
        List of files path inside parent path directory.
    """
    files_path_arr = []
    ignore_cuz_content_name_have_some_of_the_words = False
    for content_name in content_arr:
        
        content_path = f'{parent_path}/{content_name}'
        
        ignore_words_len = len(ignore_if_contain_some_of_the_words)
        for index1, word in enumerate(ignore_if_contain_some_of_the_words):
            if word in content_name:
                ignore_cuz_content_name_have_some_of_the_words = True
                break
            elif index1 == ignore_words_len - 1:
                ignore_cuz_content_name_have_some_of_the_words = False
                
        if os.path.isfile(content_path):
            if content_name not in ignore_files \
            and not ignore_cuz_content_name_have_some_of_the_words:
                files_path_arr.append(
                    join_path(parent_path, content_name)
                )
                
        elif content_name not in ignore_folders \
        and not ignore_cuz_content_name_have_some_of_the_words:
            next_parent_path = join_path(parent_path, content_name)
            next_content_arr = os.listdir(next_parent_path)
            files_path_arr = files_path_arr \
                + get_files_path(
                    next_parent_path, 
                    next_content_arr, 
                    ignore_files, 
                    ignore_folders, 
                    ignore_if_contain_some_of_the_words
                )
    return files_path_arr


def list_path_of_all_files_inside_directory(
    root_directory_path: str,
    ignore_files: List[str]=[],
    ignore_folders: List[str]=[],
    ignore_if_contain_some_of_the_words: List[str]=[]
) -> List:
    """Return array containing all abs path inside some directory.

    Parameters
    ----------
    root_directory_path : str
        Directory abs path. Search for files inside this directory
    ignore_files : List[str], optional
        Ignore files if name equal some of this list, by default []
    ignore_folders : List[str], optional
        Ignore folders if name equal some of this list, by default []
    ignore_if_contain_some_of_the_words : List[str], optional
        Ignore folders and files if name cotain some of 
        the names presented on this list, by default []

    Returns
    -------
    List
        List of files path inside parent path directory.
    """
    try:
      dataDirectoryContent = os.listdir(root_directory_path)
    except:
      dataDirectoryContent = []
    itemsPath = []
    if len(dataDirectoryContent) > 0:
      itemsPath = get_files_path(
            parent_path=root_directory_path, 
            content_arr=dataDirectoryContent, 
            ignore_files=ignore_files, 
            ignore_folders=ignore_folders, 
            ignore_if_contain_some_of_the_words=ignore_if_contain_some_of_the_words
        )
    return itemsPath


def search_by_name_on_directory(
    name: str, 
    directory: str,
) -> List[str]:
    """Search files that contain substring on name inside directory

    Parameters
    ----------
    name : str
        Name of file that will be searched
    dir : str
        Directory to search for file

    Returns
    -------
    List[str]
        All file names that was found
    """
    content = os.listdir(directory)
    result = [filename for filename in content if name in filename]
    return result
    
    
def create_directories_of_path(
    path: str
) -> None:
    """Create all directories of path.

    Parameters
    ----------
    path : str
        Desired path.
        
    Returns
    -----
    None
        None is returned
    """
    
    split_path = None
    
    if get_os_name() == "Windows": 
        split_path = path.split("\\")
    elif get_os_name() == "Linux": 
        split_path = path.split("/")
    elif get_os_name() == "Darwin": 
        split_path = path.split("\\")
        
    temp_path = None
    
    for index, env_path in enumerate(split_path):
        if not env_path: 
            continue
        if len(env_path.split(".")) > 1 and index == len(split_path) - 1: 
            break
        if temp_path == None: 
            if get_os_name() == "Windows": 
                temp_path = env_path + '\\'
                continue
            elif get_os_name() == "Linux": 
                temp_path = "/"
        temp_path = join_path(temp_path, env_path)
        if not os.path.exists(temp_path): 
            try:
                os.makedirs(temp_path, exist_ok=True)
            except Exception as e:
                template = "An exception of type {0} occurred. Arguments:\n{1!r}"
                message = template.format(type(e).__name__, e.args)
                print("Error: " + str(e))
                print(message)


def save_json(
    data: any, 
    path: str,
    indent: int=4,
    sort_keys: bool=False
) -> None:
    """Save JSON file.

    Parameters
    ----------
    data : any
        Json data
    path : str
        Destination path of file that will be saved
    indent : int, optional
        Indentation of JSON, by default 4
    sort_keys : bool, optional
        If keys of JSON will be sorted, by default False
    
    Returns
    ------
    None
        None is returned
        
    Raises
    ------
    Exception
        If path already exists.
    """
    
    if os.path.exists(path):
        raise Exception('Path already exists.')
    
    create_directories_of_path(path)
    
    if path.endswith('.gz'):
        with gzip.open(path, 'wt', encoding="ascii") as zipfile:
            json.dump(data, zipfile)
    else:
        with open(path, 'w') as f:
            json.dump(data, f, indent=indent, sort_keys=sort_keys)
                
        
def load_json(
    path: str,
) -> any:
    """Load JSON file.

    Parameters
    ----------
    path : str
        Path of file that will be loaded.
    
    Returns
    ------
    None
        None is returned
        
    Raises
    ------
    Exception
        If path does not exists.
    """
    if not os.path.exists(path):
        raise Exception('Path does not exists.')
    methods_by_type_of_file = {
        '.gz': {'open': gzip.open,'apply_on_data': lambda data: data.decode('utf-8')},
        'others': {'open': open,'apply_on_data': lambda data: data},
    }
    extension_path = '.' + path.split('.')[-1]
    methods = methods_by_type_of_file.get(
        extension_path, methods_by_type_of_file['others']
    )
    open_method = methods['open']
    apply_on_data_method = methods['apply_on_data']
    try:
        with open_method(path, "r") as f:
            data = f.read()
            data = json.loads(
                apply_on_data_method(data)
            )
    except json.JSONDecodeError as exc:
        data = []
        for line in open_method(path, 'r'):
            try:
                row = json.loads(line)
                data.append(row)
            except Exception as exc:
                pass
    return data
    
    
def rm_file(
    path: str, 
    ignore_if_path_not_exists: bool=False
) -> None: 
    """Delete file

    Parameters
    ----------
    path : str
        Abs path of file
    ignore_if_path_not_exists : bool, optional
        If True, if path not exists, do not raise Exception, 
        if False, if path not exists, raise Exception, 
        by default False
    """
    if ignore_if_path_not_exists:
      if os.path.exists(path):
        os.remove(path)
    else:
      os.remove(path)


""" 
This module was created to define utils functions that can be used
in many situations.

"""

import os
from datetime import datetime
import re
import platform
from types import FunctionType, LambdaType
from typing import Dict, List, Tuple, Union, Optional, TypedDict
import json
import numpy as np
import pandas as pd
from pandas import DataFrame


def progress(
    current: int, 
    total: int, 
) -> int:
    """Get percentage of progress of some work

    Parameters
    ----------
    current : int
        Current number of points that was 
        be done until this moment.
    total : int
        Toal number of progress points

    Returns
    -------
    int
        The percentage between 0 and 100
    """
    progress = int(round(current * 100 / total, 0))
    if progress >= 100 \
        or (progress != 100 and current == total): 
        progress = 100
    return progress


def flatten(
    dict_: Dict, 
    separator: str='.',
    flatten_list: bool=False
) -> Dict:
    """Flatten dict.

    Parameters
    ----------
    dict_ : Dict
        Ordinary Dict to be flattening
    separator : str, optional
        Separator of dict depth keys when flattening, by default '.'

    Returns
    -------
    Dict
        The flatten dict
    """
    def _flatten(_dict_: Dict, _separator: str, parent_key: str) -> Dict:
        items = []
        for k, v in _dict_.items():
            new_key = parent_key + _separator + k if parent_key else k
            if isinstance(v, dict):
                items.extend(
                    _flatten(v, _separator=_separator, parent_key=new_key).items()
                )
            else:
                if isinstance(v, list):
                    new_v = []
                    for idx, item in enumerate(v):
                        if isinstance(item, dict):
                            val = _flatten(item, _separator, parent_key='')
                        else:
                            val = item
                            
                        if flatten_list:
                            items.append((f'{new_key}{_separator}$idx:{idx}', val))
                        else:
                            new_v.append(val)
                    if not flatten_list:
                        items.append((new_key, new_v))
                else:
                    items.append((new_key, v))
        return dict(items)
    result = _flatten(dict_, separator, parent_key='')
    if flatten_list:
        result = _flatten(result, separator, parent_key='')
    return result
  
  
def re_flatten(
    dict_: Dict, 
    separator: str = '.'
) -> Dict:
    """Re flatten Dict after flatten operation has been applied to dict

    Parameters
    ----------
    dict_ : Dict
        Flatten dict
    separator : str, optional
        Key separator when flatten operation was applied, by default '.'

    Returns
    -------
    Dict
        The re-flatten dict
    """
    def _re_flatten(_dict_: Dict, _separator: str, depth: int=0) -> Dict:
        res_obj = {}
        again = False
        for key in _dict_:
            key_splited = key.split(_separator)
            if len(key_splited) > 1:
                l1 = _separator.join(key_splited[:-1])
                l2 = key_splited[-1]
                res_obj[l1] = res_obj.get(l1, {}) if depth == 0 else _dict_.get(l1, {})
                res_obj[l1][l2] = _dict_[key]
                again = True
            else:
                val = _dict_[key]
                """ if type(val) == list:
                    val = []
                    for item in _dict_[key]:
                        if type(item) == dict:
                            val.append(_re_flatten(item, _separator, depth=0))
                        else:
                            val.append(item) """
                res_obj[key] = val
        if again:
            res_obj = _re_flatten(res_obj, _separator, depth=depth + 1)
        return res_obj
    result = _re_flatten(dict_, separator, depth=0)
    return result


def is_float(
    my_str: str
) -> bool:
    """Verify if string is float format.

    Parameters
    ----------
    my_str : str
        Input string

    Returns
    -------
    bool
        True if string is float, else, False
    """
    
    resp = False
    try:
        float(my_str)
        resp = True
    except Exception:
        resp = False
    return resp


def is_int(
    my_str: str
) -> bool:
    """Verify if string is int format.

    Parameters
    ----------
    my_str : str
        Input string

    Returns
    -------
    bool
        True if string is int, else, False
    """
    
    resp = False
    try:
        int(my_str)
        resp = True
    except Exception:
        resp = False
    return resp


def list_ops(
    list1: List,
    list2: List,
    ops: str='difference'
) -> List:
    """Realize operation between two lists.
    
    Difference:
    - Get element which exists in 'list1' but not exist in 'list2'.

    Parameters
    ----------
    list1 : List
        List one.
    list2 : List
        List two.
    ops : str
        The desired operation to be performed 
        on list, by default 'difference'.

    Returns
    -------
    List
        Returns the result of the selected operation.
    """
    
    if ops == 'difference':
        s = set(list2);
        return [x for x in list1 if x not in s]
    elif ops == 'intersection':
        s = set(list2);
        return [x for x in list1 if x in s]
    elif ops == 'union':
        return list_ops(list1, list2, ops='difference') + list2
    else:
        raise Exception('Invalid "ops" option.')

# *****************************
# [DEPRECATED]
# *****************************
List1 = int
List2 = int
def list_difference(
    list1: List,
    list2: List,
    axis: Union[List1, List2]=1
) -> List:
    """Get element which exists in 'list1' but not exist in 'list2' and
    get elements which exists in 'list1' but not exist in 'list2'.

    Parameters
    ----------
    list1 : List
        List one.
    list2 : List
        List two. 
    axis : Union[List1, List2], optional
        Can be only 1 or 2. 
        If 1, return values that is only in list 1,
        if 2, return values that is only in list 2,
        by default 1

    Returns
    -------
    ListDifference
        Returns the elements which exists in 'list1' and not exists 
        in 'list2' and elements which exists in 'list1' and not exists 
        in 'list2'
    """
    
    if axis == 1:
        return [item for item in list1 if item not in list2]
    elif axis == 2:
        return [item for item in list2 if item not in list1]
    else:
        raise Exception('Invalid "axis" option.')


def camel_to_snake(
    name: str
) -> str:
    """Transform camel case to snake case

    Parameters
    ----------
    name : str
        Camel case name

    Returns
    -------
    str
        Snake case name
    """
    name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()


def to_title(
    text: str,
    upper_initials: bool=False
) -> str:
    """Transforms text to title
    
    Title remains first letter in capslock and lower for the remaining. 
    Replaces '_' by ' '. If in camel case, add space.

    Parameters
    ----------
    text : str
        Text to be transformed in title
    upper_initials : bool, optional
        If True, initials of all words will be in Caps Lock, 
        If False, only the first letter of the received string will be in Caps Lock, 
        by default False.

    Returns
    -------
    str
        Title text
    """
    text = camel_to_snake(text)
    text = text.replace('_', ' ')
    if not upper_initials:
        return text[0].upper() + text[1:]
    else:
        text = text.split(' ')
        text = [x[0].upper() + x[1:] for x in text]
        text = ' '.join(text)
        return text
    

def hightlight_header(
    title: str,
    break_line_after: int=1
) -> str:
    """Generate hightlight title to print in console

    Parameters
    ----------
    title : str
        Text on header
    break_line_after : int, optional
        Num of break lines to separe title from text, by default 1

    Returns
    -------
    str
        Hightlighted header
    """
    real_title = f'!*** {title} ***!'
    hightlight = f'!{"*" * (len(real_title) - 2)}!'
    return f'{hightlight}\n{real_title}\n{hightlight}' + ("\n" * break_line_after)


# *******************************************
# deprecated ********************************
# *******************************************
def json_pp_message(
    dict_data: dict, 
    sort_keys: bool=True, 
    indent: int=4
) -> str:
    """Get json in a pretty format

    Parameters
    ----------
    dict_data : dict
        Raw data
    sort_keys : bool, optional
        Sort keys of dict, by default True
    indent : int, optional
        Indent desired in message, by default 4

    Returns
    -------
    str
        Json message with pretty format
    """
    return json.dumps(dict_data, sort_keys = sort_keys, indent = indent)


# *******************************************
# deprecated ********************************
# *******************************************
def json_pp(
    dict_data: dict,
    sort_keys: bool=False,
    indent: int=4
) -> None:
    """Pretty print of json data

    Parameters
    ----------
    dict_data : dict
        Raw json data
    sort_keys : bool, optional
        Sort keys of json, by default True
    indent : int, optional
        Indentation of json message, by default 4
    """
    print(
        json_pp_message(
            dict_data=dict_data,
            sort_keys=sort_keys,
            indent=indent,
        )
    )
    
    
def get_os_name() -> str:
    """Get operational system

    Returns
    -------
    str
        Name of operational system
    """
    return platform.system()
    
    
def join_path(
    *args: Tuple[str],
) -> str:
    """Join multiple paths.
    
    Consider special cases, like when one of the paths
    starts with '/'.

    Returns
    -------
    str
        Joined path.
    """
    path = None
    for index, arg in enumerate(args):
        if index != 0 and arg and arg[0] == "/":
            arg = arg[1:]
        if index == 0: 
            path = os.path.abspath(os.path.join(arg))
        else: 
            path = os.path.abspath(os.path.join(path, arg))
    return path


def display_time(
    seconds: int, 
    granularity: int=2
) -> str:
    """Display time based on granularity by converting seconds.
    
    Convert seconds to weeks, days, hours, minutes and seconds.
    
    Parameters
    ----------
    seconds : int
        Number of seconds.
    granularity : int, optional
        Granularity of response, 
        by default 2.

    Returns
    -------
    str
        The corresponding time based on granularity.
    """
    
    intervals = (
        ('months', 604800),  # 60 * 60 * 24 * 30
        ('weeks', 604800),  # 60 * 60 * 24 * 7
        ('days', 86400),    # 60 * 60 * 24
        ('hours', 3600),    # 60 * 60
        ('minutes', 60),
        ('seconds', 1),
    )
    
    result = []

    for name, count in intervals:
        value = seconds // count
        if value:
            seconds -= value * count
            if value == 1:
                name = name.rstrip('s')
            result.append("{} {}".format(value, name))
    return ', '.join(result[:granularity])


def isnan(
    value: any
) -> bool:
    """Check if inputed value is nan.

    Parameters
    ----------
    value : any
        Value to check if is nan.

    Returns
    -------
    bool
        If is nan, returns True, 
        else, returns False
    """
    resp = False
    has_except = False
    
    try:
        resp = pd.isna(value)
    except:
        resp = False
        has_except = True
        
    if has_except:
        has_except = False
        try:
            resp = np.isnan(value)
        except:
            resp = False
            has_except = True
            
    if has_except:
        has_except = False
        resp = False
        
    return resp


def iso_str(time_sep='-') -> str:
    """Get time as string. Useful to generate unique file names

    Parameters
    ----------
    time_sep : str, optional
        Separtor of hours, minutes, seconds and 
        microseconds, by default '-'

    Returns
    -------
    str
        String representing now date.
    """
    return datetime.now().strftime(f"%Y-%m-%dT%H:%M:%S.%fZ")


class ValueRestrictionsAsJson(TypedDict):
    bigger_than: Optional[float]
    bigger_or_equal_than: Optional[float]
    smaller_than: Optional[float]
    smaller_or_equal_than: Optional[float]
    equals_to: Optional[float]
    not_equals_to: Optional[float]


class ValueRestrictions:
    """Handle restrictions in values, especially useful when restrict 
    DataFrame numerical values.
    """

    @staticmethod
    def BIGGER_THAN(desired_compare_value, reference_value): 
        return desired_compare_value > reference_value

    @staticmethod
    def BIGGER_OR_EQUAL_THAN(desired_compare_value, reference_value): 
        return desired_compare_value >= reference_value

    @staticmethod
    def SMALLER_THAN(desired_compare_value, reference_value): 
        return desired_compare_value < reference_value

    @staticmethod
    def SMALLER_OR_EQUAL_THAN(desired_compare_value, reference_value): 
        return desired_compare_value <= reference_value

    @staticmethod
    def EQUALS_TO(desired_compare_value, reference_value): 
        return desired_compare_value == reference_value

    @staticmethod
    def NOT_EQUALS_TO(desired_compare_value, reference_value): 
        return desired_compare_value != reference_value
    
    def __init__(
        self,
        bigger_than: float=None,
        bigger_or_equal_than: float=None,
        smaller_than: float=None,
        smaller_or_equal_than: float=None,
        equals_to: float=None,
        not_equals_to: float=None,
        extra_restrictions: List[Union[FunctionType, LambdaType]]=None
    ) -> None:
        self.bigger_than = bigger_than
        self.bigger_or_equal_than = bigger_or_equal_than
        self.smaller_than = smaller_than
        self.smaller_or_equal_than = smaller_or_equal_than
        self.equals_to = equals_to
        self.not_equals_to = not_equals_to
        self.extra_restrictions = extra_restrictions

    def as_json(self) -> dict:
        resp = {}
        if self.bigger_than != None: 
            resp["bigger_than"] = self.bigger_than
        if self.bigger_or_equal_than != None: 
            resp["bigger_or_equal_than"] = self.bigger_or_equal_than
        if self.smaller_than != None: 
            resp["smaller_than"] = self.smaller_than
        if self.smaller_or_equal_than != None: 
            resp["smaller_or_equal_than"] = self.smaller_or_equal_than
        if self.equals_to != None: 
            resp["equals_to"] = self.equals_to
        if self.not_equals_to != None: 
            resp["not_equals_to"] = self.not_equals_to
        return resp

    def verify(
        self,
        value
    ) -> bool:
        resp = []
        if self.bigger_than != None: 
            resp.append(self.BIGGER_THAN(value, self.bigger_than))
        if self.bigger_or_equal_than != None: 
            resp.append(self.BIGGER_OR_EQUAL_THAN(value, self.bigger_or_equal_than))
        if self.smaller_than != None: 
            resp.append(self.SMALLER_THAN(value, self.smaller_than))
        if self.smaller_or_equal_than != None: 
            resp.append(self.SMALLER_OR_EQUAL_THAN(value, self.smaller_or_equal_than))
        if self.equals_to != None: 
            resp.append(self.EQUALS_TO(value, self.equals_to))
        if self.not_equals_to != None: 
            resp.append(self.NOT_EQUALS_TO(value, self.not_equals_to))
        resp = [x for x in resp if not x]
        if len(resp) > 0: 
            resp = False
        else: 
            resp = True

        if self.extra_restrictions != None:
            extra_resp = []
            for extra_restriction in self.extra_restrictions:
                extra_resp.append(extra_restriction(value)) 
            extra_resp = [x for x in extra_resp if not x]
            if len(extra_resp) > 0: extra_resp = False
            else: extra_resp = True
            resp = resp and extra_resp

        return resp
    
    def restrict_df(
        self,
        df: DataFrame,
        cols: Union[str, List[str]],
    ) -> DataFrame:
        
        real_cols = cols
        if type(cols) == str:
            real_cols = [cols]
        
        if self.bigger_than is not None: 
            df = df[np.all(self.BIGGER_THAN(df[real_cols], self.bigger_than), axis=1)]
        if self.bigger_or_equal_than is not None: 
            df = df[np.all(self.BIGGER_OR_EQUAL_THAN(df[real_cols], self.bigger_or_equal_than), axis=1)]
        if self.smaller_than is not None: 
            df = df[np.all(self.SMALLER_THAN(df[real_cols], self.smaller_than), axis=1)]
        if self.smaller_or_equal_than is not None: 
            df = df[np.all(self.SMALLER_OR_EQUAL_THAN(df[real_cols], self.smaller_or_equal_than), axis=1)]
        if self.equals_to is not None: 
            df = df[np.all(self.EQUALS_TO(df[real_cols], self.equals_to), axis=1)]
        if self.not_equals_to is not None: 
            df = df[np.all(self.NOT_EQUALS_TO(df[real_cols], self.not_equals_to), axis=1)]
        return df
    
    def restrict_numpy(
        self,
        matrix: np.array,
        cols_indice: Union[int, List[int]],
    ) -> DataFrame:
        
        real_cols_indice = cols_indice
        if type(cols_indice) == int:
            cols_indice = [cols_indice]
        
        if self.bigger_than is not None: 
            # if is_int:
            #     matrix = matrix[
            #         np.nonzero(self.BIGGER_THAN(matrix[:, col_indice], self.bigger_than))
            #     ]
            # else:
            matrix = matrix[
                np.nonzero(np.all(
                    self.BIGGER_THAN(matrix[:, real_cols_indice], self.bigger_than),
                    axis=1
                ))
            ]                
        if self.bigger_or_equal_than is not None: 
            # if is_int:
            #     matrix = matrix[
            #         np.nonzero(self.BIGGER_OR_EQUAL_THAN(matrix[:, col_indice], self.bigger_or_equal_than))
            #     ]
            # else:
            matrix = matrix[
                np.nonzero(np.all(
                    self.BIGGER_OR_EQUAL_THAN(matrix[:, real_cols_indice], self.bigger_or_equal_than),
                    axis=1
                ))
            ]
        if self.smaller_than is not None: 
            # if is_int:
            #     matrix = matrix[
            #         np.nonzero(self.SMALLER_THAN(matrix[:, col_indice], self.smaller_than))
            #     ]
            # else:
            matrix = matrix[
                np.nonzero(np.all(
                    self.SMALLER_THAN(matrix[:, real_cols_indice], self.smaller_than),
                    axis=1
                ))
            ]
        if self.smaller_or_equal_than is not None: 
            # if is_int:
            #     matrix = matrix[
            #         np.nonzero(self.SMALLER_OR_EQUAL_THAN(matrix[:, col_indice], self.smaller_or_equal_than))
            #     ]
            # else:
            matrix = matrix[
                np.nonzero(np.all(
                    self.SMALLER_OR_EQUAL_THAN(matrix[:, real_cols_indice], self.smaller_or_equal_than),
                    axis=1
                ))
            ]
        if self.equals_to is not None: 
            # if is_int:
            #     matrix = matrix[
            #         np.nonzero(self.EQUALS_TO(matrix[:, col_indice], self.equals_to))
            #     ]
            # else:
            matrix = matrix[
                np.nonzero(np.all(
                    self.EQUALS_TO(matrix[:, real_cols_indice], self.equals_to),
                    axis=1
                ))
            ]
        if self.not_equals_to is not None: 
            # if is_int:
            #     matrix = matrix[
            #         np.nonzero(self.NOT_EQUALS_TO(matrix[:, col_indice], self.not_equals_to))
            #     ]
            # else:
            matrix = matrix[
                np.nonzero(np.all(
                    self.NOT_EQUALS_TO(matrix[:, real_cols_indice], self.not_equals_to),
                    axis=1
                ))
            ]
        return matrix

    def initialize_from_dict(
        self,
        value_restrictions_dict: ValueRestrictionsAsJson    
    ) -> None:
        self.bigger_than = value_restrictions_dict.get('bigger_than', None)
        self.bigger_or_equal_than = value_restrictions_dict.get('bigger_or_equal_than', None)
        self.smaller_than = value_restrictions_dict.get('smaller_than', None)
        self.smaller_or_equal_than = value_restrictions_dict.get('smaller_or_equal_than', None)
        self.equals_to = value_restrictions_dict.get('equals_to', None)
        self.not_equals_to = value_restrictions_dict.get('not_equals_to', None)
        self.extra_restrictions = value_restrictions_dict.get('extra_restrictions', None)


""" 
This module was created to handle Features construction and 
other stuff related to features from Machine Learn Model.

"""


from drtools.utils import list_ops
from pandas import DataFrame
import pandas as pd
from typing import List, Union, Dict, TypedDict


ColumnName = str
EncodeValue = List[Union[str, int]]
class EncondeOptions(TypedDict):
    EncodeValues: List[EncodeValue]
    DropRedundantColVal: str


def single_ohe(
    dataframe: DataFrame,
    column: str,
    encode_values: List[EncodeValue],
    prefix: str=None,
    prefix_sep: str="_",
    drop_self_col: bool=True,
    drop_redundant_col_val: str=None
) -> DataFrame:
    """One hot encode one column, drop original column after 
    generate encoded and drop dummy cols that is not desired on 
    final data.
    
    Parameters
    ----------
    dataframe : DataFrame
        DataFrame containing data to encode.
    column : str
        Name of column to one hot encode.
    encode_values: List[Union[str, int]]
        List of values to encode.
    prefix: str, optional
        Prefix of encoded column. If None, 
        the prefix will be the column name, by default None.
    prefix_sep: str, optional
        Separation string of Prefix and Encoded Value, 
        by default "_".
    drop_self_col: bool, optional
        If True, the encoded column will be deleted. 
        If False, the encoded column will not be deleted, 
        by default True.
    drop_redundant_col_val: str, optional
        If is not None, supply value that will corresnponde to encode column and 
        the encoded column will be dropped after generate encoded columns, 
        by default None.
        
    Returns
    -------
    DataFrame
        The DataFrame containing encoded columns.
    """
    if prefix is None:
        prefix = column    
    finals_ohe_cols = [
        f'{prefix}{prefix_sep}{x}'
        for x in encode_values
    ]
    df = dataframe.copy()
    dummies = pd.get_dummies(df[column], prefix=prefix, prefix_sep= prefix_sep)
    drop_cols = list_ops(dummies.columns, finals_ohe_cols)
    df = pd.concat([df, dummies], axis=1)
    if drop_self_col:
        drop_cols = drop_cols + [column]
    df = df.drop(drop_cols, axis=1)
    # insert feature that not has on received dataframe
    for col in finals_ohe_cols:
        if col not in df.columns:
            df[col] = 0
    if drop_redundant_col_val is not None:
        drop_encoded_col_name = f'{prefix}{prefix_sep}{drop_redundant_col_val}'
        if drop_encoded_col_name in df.columns:
            df = df.drop(drop_encoded_col_name, axis=1)
    return df


def one_hot_encoding(
    dataframe: DataFrame,
    encode: Dict[ColumnName, EncondeOptions],
    prefix: str=None,
    prefix_sep: str="_",
    drop_self_col: bool=True,
) -> DataFrame:
    """One hot encode variables, drop original column that 
    generate encoded and drop dummy cols that is not present 
    on the input features.
    
    Parameters
    ----------
    dataframe : DataFrame
        DataFrame containing data to encode.
    encode : Dict[ColumnName, List[EncodeValue]]
        The dict containing column names and values to encode.
    prefix: str, optional
        Prefix of encoded column. If None, 
        the prefix will be the column name, by default None.
    prefix_sep: str, optional
        Separation string of Prefix and Encoded Value, 
        by default "_".
    drop_self_col: bool, optional
        If True, the encoded column will be deleted. 
        If False, the encoded column will not be deleted, 
        by default True.
        
    Returns
    -------
    DataFrame
        The DataFrame containing encoded columns.
    """
    df = dataframe.copy()
    for column_name, encode_options in encode.items():
        encode_values = encode_options['EncodeValues']
        drop_redundant_col_val = encode_options.get('DropRedundantColVal', None)
        df = single_ohe(
            dataframe=df,
            column=column_name,
            encode_values=encode_values,
            drop_redundant_col_val=drop_redundant_col_val,
            prefix=prefix,
            prefix_sep=prefix_sep,
            drop_self_col=drop_self_col,
        )
    return df




import unittest
from drtools.utils import (
    flatten, 
    re_flatten, 
    progress
)


class TestProgress(unittest.TestCase):

    def test_current_gt_total(self):
        observed = (51, 50)
        expected = 100
        response = progress(*observed)
        self.assertEqual(response, expected)
        
    def test_current_eq_total(self):
        observed = (50, 50)
        expected = 100 
        response = progress(*observed)
        self.assertEqual(response, expected)

    def test_current_lt_total(self):
        observed = (40, 50)
        expected = 80
        response = progress(*observed)
        self.assertEqual(response, expected)


class TestFlatten(unittest.TestCase):

    def test_ordinary(self):
        observed = {'a': 1, 'b': 2, 'c': {'d': 3, 'e': {'f': 4, 'g': {'h': 5}}}}
        expected = {'a': 1, 'b': 2, 'c.d': 3, 'c.e.f': 4, 'c.e.g.h': 5}
        response = flatten(observed)
        self.assertEqual(response, expected)
        
    def test_with_list(self):
        observed = {'a': 1, 'b': 2, 'c': {'d': 3, 'e': {'f': [1,2,3], 'g': {'h': 5}}}}
        expected = {'a': 1, 'b': 2, 'c.d': 3, 'c.e.f': [1,2,3], 'c.e.g.h': 5}
        response = flatten(observed)
        self.assertEqual(response, expected)

    def test_with_dict_inside_list(self):
        observed = {'a': 1, 'b': 2, 'c': {'d': 3, 'e': {'f': [1,2,{'k1': 1, 'k2': {'k3': 123, 'k4': {'k5': 'asdasd'}}}], 'g': {'h': 5}}}}
        expected = {'a': 1, 'b': 2, 'c.d': 3, 'c.e.f': [
            1, 2,
            {'k1': 1, 'k2.k3': 123, 'k2.k4.k5': 'asdasd'}
        ], 'c.e.g.h': 5}
        response = flatten(observed)
        self.assertEqual(response, expected)
        
        
class TestReFlatten(unittest.TestCase):

    def test_ordinary1(self):
        observed = {'a': 1, 'b': 2, 'c.d': 3, 'c.f': 4}
        expected = {'a': 1, 'b': 2, 'c': {'d': 3, 'f': 4}}
        response = re_flatten(observed)
        self.assertEqual(response, expected)
        
    def test_ordinary2(self):
        observed = {'a': 1, 'b': 2, 'c.d': 3, 'c.e.f': 4, 'c.e.g.h': 5}
        expected = {'a': 1, 'b': 2, 'c': {'d': 3, 'e': {'f': 4, 'g': {'h': 5}}}}
        response = re_flatten(observed)
        self.assertEqual(response, expected)
        
    def test_with_list(self):
        observed = {'a': 1, 'b': 2, 'c.d': 3, 'c.e.f': [1,2,3], 'c.e.g.h': 5}
        expected = {'a': 1, 'b': 2, 'c': {'d': 3, 'e': {'f': [1,2,3], 'g': {'h': 5}}}}
        response = re_flatten(observed)
        self.assertEqual(response, expected)

    """ def test_with_dict_inside_list(self):
        observed = {'a': 1, 'b': 2, 'c.d': 3, 'c.e.f': [
            1, 2,
            {'k1': 1, 'k2.k3': 123, 'k2.k4.k5': 'asdasd'}
        ], 'c.e.g.h': 5}
        expected = {'a': 1, 'b': 2, 'c': {'d': 3, 'e': {'f': [1,2,{'k1': 1, 'k2': {'k3': 123, 'k4': {'k5': 'asdasd'}}}], 'g': {'h': 5}}}}
        response = re_flatten(observed)
        self.assertEqual(response, expected) """

if __name__ == '__main__':
    unittest.main()


""" 
This module was created to handle
Driver initialization.

"""


from typing import Union, List
from selenium import webdriver
import logging
from selenium.webdriver.chrome.service import Service as ChromeService
from webdriver_manager.chrome import ChromeDriverManager
from webdriver_manager.core.utils import ChromeType
from selenium.webdriver.chrome.webdriver import WebDriver


GOOGLE = 'GOOGLE'
CHROMIUM = 'CHROMIUM'
BRAVE = 'BRAVE'
MSEDGE = 'MSEDGE'


def chrome_start(
    chrome_type: Union[
        GOOGLE, CHROMIUM, BRAVE, MSEDGE
    ]=GOOGLE,
    chrome_options_arguments: List[str]=[],
    warning_logs: bool=True,
    load_images: bool=False,
    load_js: bool=True
) -> WebDriver:
    """Start Selenium Chrome Driver.
    
    Example
    -------
    Examples of Chrome Options Arguments usage:
    
    **Remove UI**
    
    - chrome_options.add_argument("--headless")
    - chrome_options.add_argument("--no-sandbox")
    - chrome_options.add_argument("--mute-audio")    
    
    **Change window size**
    
    - chrome_options.add_argument("--start-maximized")
    - chrome_options.add_argument("--window-size=1920x1080")    
    
    **Change default download location**
    
    - chrome_options.add_argument("download.default_directory=C:/Downloads")
    
    Parameters
    ----------
    chrome_type : Union[GOOGLE, CHROMIUM, BRAVE, MSEDGE]
        The chrome type will be loaded, by default GOOGLE
    chrome_options_arguments : List[str]
        A list of chrome options arguments, by default []
    warning_logs : bool
        Display logs of selenium webdriver and urllib3 in
        warning level, by default True
    load_images : bool
        If images of page should be loaded, by default False    
    load_js : bool
        If Javascript of page should be loaded, by default True
        
    Returns
    -------
    WebDriver
        Selenium WebDriver
    """
    
    # Only display possible problems
    if warning_logs:
        logging.getLogger('selenium.webdriver.remote.remote_connection') \
            .setLevel(logging.WARNING)
        logging.getLogger('urllib3.connectionpool') \
            .setLevel(logging.WARNING)

    chrome_options = webdriver.ChromeOptions()

    # chrome options arguments
    for arg in chrome_options_arguments:
        chrome_options.add_argument(arg)

    ### This blocks images and javascript requests
    chrome_prefs = { "profile.default_content_setting_values": {} }
    
    if not load_images:
        chrome_prefs['profile.default_content_setting_values']['images'] = 2
    if not load_js:
        chrome_prefs['profile.default_content_setting_values']['javascript'] = 2

    chrome_options.experimental_options["prefs"] = chrome_prefs
    ###

    driver = webdriver.Chrome(
        service=ChromeService(ChromeDriverManager(
            chrome_type=getattr(ChromeType, chrome_type)
            ).install()),
        chrome_options=chrome_options
    )
    
    return driver


""" 
This module was created to handle common
geo data and issues.

"""


from typing import Tuple, Union
from numpy import inf
import numpy as np
import geopy.distance


Lower = 'lower'
Upper = 'upper'
Latitude = 'lat'
Longitude = 'lon'


def distance_by_geodesic_coord(
    p1: Tuple[float, float],
    p2: Tuple[float, float],
) -> float:
    """Get distance in meters between two points by geodesic coordinates.

    Parameters
    ----------
    p1 : Tuple[float, float]
        Coordinates of point 1 in format (Latitude, Longitude).
    p2 : Tuple[float, float]
        Coordinates of point 2 in format (Latitude, Longitude).

    Returns
    -------
    float
        The distance in meters between two points by geodesic coordinates. 
    """
    return geopy.distance.geodesic(p1, p2).m


def find_y_value(
    *args, **kwargs
) -> any:
    interval_arr = args[0]
    geodesic_axis = args[1]
    if geodesic_axis == 'lat':
        y_bot_value = geopy.distance.geodesic(
            (0, 0), (interval_arr[0], 0)
        ).m
        y_top_value = geopy.distance.geodesic(
            (0, 0), (interval_arr[1], 0)
        ).m
    elif geodesic_axis == 'lon':
        y_bot_value = geopy.distance.geodesic(
            (0, 0), (0, interval_arr[0])
        ).m
        y_top_value = geopy.distance.geodesic(
            (0, 0), (0, interval_arr[1])
        ).m
    return y_bot_value, y_top_value


def distance_to_geodesic(
    distance: float,
    init_guess_interval: Tuple[float, float]=(0, 1),
    geodesic_axis: Union[Latitude, Longitude]=Latitude,
    precision: float=1,
    boundary: Union[Lower, Upper]=Lower
) -> Tuple[
    float, 
    float, 
    Tuple[float, float],
    int
]:
    """Tranform distance in meters to corresponding latitude or 
    longitude angle value.

    Parameters
    ----------
    distance : float
        Distance in meters
    init_guess_interval : Tuple[float, float], optional
        Guess interval where corresponding value will 
        be, by default (0, 1)
    geodesic_axis : Union[Latitude, Longitude], optional
        Value of angle on Latitude or Longitude axis, by default Latitude
    precision : float, optional
        Precision of angle value, by default 1
    boundary : Union[Lower, Upper], optional
        If Lower, returns value of angle corresponding
        to some distance smaller than **distance**. If 
        Upper, returns value of angle corresponding
        to some distance bigger than **distance**, 
        by default Lower

    Returns
    -------
    Tuple[ float, float, Tuple[float, float], int ]
        Returns the distance corresponding to desired angle, 
        the angle value, the interval containing the real angle value 
        and the number of iterations until find the value of angle
    """
    
    y_value = inf
    interval_arr = np.array(init_guess_interval)
    
    boundary_condition = lambda y, val: \
        y > val if boundary == Lower else \
        y < val if boundary == Upper else \
        False
    
    iters = 0
    while abs(y_value - distance) > precision or boundary_condition(y_value, distance):
        
        # find y value
        y_bot_value, y_top_value = find_y_value(
            interval_arr, geodesic_axis
        )
        
        # verify inequality between value and y value
        if y_bot_value <= distance <= y_top_value:
            interval_mean = interval_arr.mean()
            y_mean_value, _ = find_y_value(
                np.array([interval_mean, interval_mean]), geodesic_axis
            )
            if y_mean_value > distance:
                interval_arr = np.array([
                    interval_arr[0], interval_mean
                ])
            elif y_mean_value < distance:
                interval_arr = np.array([
                    interval_mean, interval_arr[1]
                ])
            else:
                interval_arr = np.array([
                    interval_mean, interval_mean
                ])
        else:
            interval_arr = np.array([
                interval_arr[1],
                2*interval_arr[1] - interval_arr[0],
            ])
        
        y_value = y_mean_value
        iters = iters + 1
    
    return y_value, interval_mean, tuple(interval_arr), iters


"""This module was created to handle concurrent executions

"""

import os
import concurrent
import concurrent.futures
import logging
from logging import Logger
from datetime import datetime
from typing import Dict, List
from types import FunctionType
from drtools.utils import progress, display_time
from drtools.logs import Log
import numpy as np
from numpy import inf
import math


WorkerData = any
"""any: Worker data can be any type of data."""

LambdaResponse = any
"""any: Lambda Response can be any type of data."""

class Event:
	"""Lambda event to pass to ``handle_lambda`` function

	Parameters
	----------
	parameters : any
		Parameters to pass to ``execution_function``
	execution_function : FunctionType
		The function that will be executed
	verbose : bool, optional
		If lambda execution will have logs, by default True.
	verbose_parameters_sample : bool, optional
		If True, will log only a sample of parameters, 
  		by default True.
	direct : bool, optional
		If True, pass parameters direct to execution function 
  		as arguments, by default False.
	LOGGER : Log, optional
		The Logger, by default None
	
	"""
    
	def __init__(
		self,
		parameters: any,
		execution_function: FunctionType,
		verbose: bool=True,
  		verbose_parameters_sample: bool=True,
		direct: bool=False,
  		LOGGER: Log=None
	) -> None:
		self.parameters = parameters
		self.execution_function = execution_function
		self.verbose = verbose
		self.verbose_parameters_sample = verbose_parameters_sample
		self.direct = direct
		self.LOGGER = logging if LOGGER is None else LOGGER
        

def handle_lambda(event: Event) -> LambdaResponse:
	"""Execute a function passing parameters with error handling

	Parameters
	----------
	event : Event
		The Lambda Event

	Returns
	-------
	LambdaResponse
		Lambda Response
	"""
	init = datetime.now()
	parameters = event.parameters
	execution_function = event.execution_function
	logger = event.LOGGER
	parameters_str = str(parameters)
	parameters_sample = f'{parameters_str[:100]} ... {parameters_str[-100:]}'
	if event.verbose:
		if event.verbose_parameters_sample:
			logger.info(f'Start execution with parameters (sample): {parameters_sample}')
		else:			
			logger.info(f'Start execution with parameters: {parameters}')
	function_response = None
	try:
		function_parameters = parameters if event.direct \
      		else Event(parameters=parameters, execution_function=None, LOGGER=logger)
		function_response = execution_function(function_parameters)
		if event.verbose:
			logger.info(f'Succesful execution!')
			logger.info(f'Lambda execution response:')
			response_str = str(function_response)
			log_text = f'{response_str[:100]}'
			if len(response_str) > 100:
				log_text += f' <|:::|> {response_str[-min(len(response_str) - 100, 100):]}'
			logger.info(log_text)
				
	except Exception as exc: 
		if event.verbose_parameters_sample:
			logger.error(f'Execution with parameters (sample): {parameters_sample} generate an exception: {exc}')
		else:
			logger.error(f'Execution with parameters: {parameters} generate an exception: {exc}')
	timeDiff = (datetime.now() - init).total_seconds()
	if event.verbose:
		logger.info(f"Execution ends in {timeDiff}s.")
	return function_response


class ThreadPoolExecutor:
	"""Execute function in parallel threads

	Parameters
	----------
	execution_function : FunctionType
		Function to be executed in concurrents
	worker_data : List[WorkerData]
		The input to be passed for each concurrent
		execution
	max_workers : int, optional
		Max number of parallel threads, by default 5
	worker_id_pattern : FunctionType, optional
		Function to be applied in order to uniquely
		identify a worker, by default lambdax:x
	LOGGER : Log, optional
		The Logger, by default None
	verbose_percentage : float, optional
		The percentage of workers that will
  		be logged, by default 0.1
	stop_count : int, optional
		Stop if number of processed works reach 
		this level, after that, reset counts and start 
		couting again, by default inf.
	stop_computation : FunctionType, optional
		Function to be executed when stop thread pool execution, 
  		by default None.
	del_response_when_stop : bool, optional
		If True, the current response will be deleted after 
  		stop computation executes. If False, the response will 
    	be cumulative till the end, by default False.
	verbose_parameters_sample : bool, optional
		If True, will log only a sample of parameters, 
  		by default True.
	direct : bool, optional
		If True, pass parameters direct to execution function 
  		as arguments, by default False.

	Example
	--------
	Usage example:

	>>> thread_pool_executor = ThreadPoolExecutor(...)
	>>> thread_pool_executor.start()
	>>> thread_pool_executor.get_result()
	<The execution response>
	"""

	def __init__(
		self, 
		execution_function: FunctionType, 
		worker_data: List[WorkerData], 
		max_workers: int=5, 
		worker_id_pattern: FunctionType=lambda worker: worker,
  		LOGGER: Log=None,
		verbose_percentage: float=0.1,
		stop_count: int=inf, 
		stop_computation: FunctionType=None,
  		del_response_when_stop: bool=False,
		verbose_parameters_sample: bool=False,
		direct: bool=False,
	) -> None:
		self.execution_function = execution_function 
		self.worker_data = worker_data 
		self.max_workers = max_workers 
		self.worker_id_pattern = worker_id_pattern 
		if LOGGER is not None:
			self.LOGGER = LOGGER
		else:
			self.LOGGER = Log(log_as_print=True)
		self.verbose_percentage = verbose_percentage
		self.stop_count = stop_count
		self.stop_computation = stop_computation
		self.del_response_when_stop = del_response_when_stop
		self.verbose_parameters_sample = verbose_parameters_sample
		self.direct = direct

	def start(
		self
	) -> None:
		"""Start Thread Pool Execution.
		"""
  
		self.num_of_processed_workers = 0  
		self.started_at = datetime.now()
		self.updated_at = datetime.now()
		self._progress_workers = []
		self._progress_time = []
  
		if len(self.worker_data) == 0:
			self.LOGGER.info('No data to process...')
			self.result = None
			return None	
  
		self.LOGGER.info('Starting Thread Pool Execution...')  
		thread_executor_response = self._thread_pool_executor(
			self.execution_function, 
			self.worker_data, 
			self.worker_id_pattern,
			self.max_workers,
		)  
		self.LOGGER.info('Thread Pool Execution Finished.')  
		self.result = thread_executor_response

	def _thread_pool_executor(
		self,
		execution_function, 
		worker_data: List[WorkerData], 
		worker_id_pattern: FunctionType,
		max_workers: int=5,
	) -> List[LambdaResponse]:
     
		# define verbose items
		worker_data_len = len(worker_data)
		verbose_items = np.arange(worker_data_len)
		num_elems = math.ceil(len(verbose_items) * self.verbose_percentage)
		verbose_index = np.round(np.linspace(0, worker_data_len - 1, num_elems)).astype(int)
		# verbose_index = verbose_index[1:] if len(verbose_index) > 1 else verbose_index
		verbose_index = verbose_index if verbose_index[-1] == worker_data_len - 1 \
			else verbose_index + [worker_data_len - 1]
  
		response = []
		worker_data_pack = []
		temp_data = []
		curr_count = 0

		for idx, worker in enumerate(worker_data):
			if idx > 0 and idx % self.stop_count == 0:
				worker_data_pack.append(temp_data)
				temp_data = []
			temp_data.append(worker)
		worker_data_pack.append(temp_data)
		del temp_data

       
		for idx_0, sub_worker_data in enumerate(worker_data_pack):
			curr_response = []
			with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
				futures = {
					executor.submit(
						self._track_progress, 
						{
							'event': Event(
								parameters=row, 
								execution_function=execution_function,
								verbose=(curr_count + idx) in verbose_index,
        						verbose_parameters_sample=self.verbose_parameters_sample,
        						direct=self.direct,
								LOGGER=self.LOGGER,
							),
							'current': curr_count + idx + 1,
							'total': worker_data_len
						}
					): worker_id_pattern(row) 
					for idx, row in enumerate(sub_worker_data)
				}
				for future in concurrent.futures.as_completed(futures):
					identifier_str = futures[future]
					try:
						data = future.result()
						curr_response.append(data)
					except Exception as exc:
						self.LOGGER.error(f'{identifier_str} generated an exception: {exc}')
      
			curr_count = curr_count + len(sub_worker_data)
   
			if self.stop_computation is not None:
				stop_progress_text = f'({idx_0 + 1}/{len(worker_data_pack)})'
				self.LOGGER.info(f'Stop computation execution {stop_progress_text}...')
				self.stop_computation(curr_response)
				self.LOGGER.info(f'Stop computation execution {stop_progress_text}... Done')
				self.LOGGER.info(f'Complete execution of {curr_count} from {worker_data_len} workers')    
				if self.del_response_when_stop:
					response = []
    
			response = response + curr_response
    
		return response

	def _track_progress(
     	self,	
		data: Dict
	) -> any:

		event = data.get('event')
		lambda_response = handle_lambda(event)
		# current = data.get('current')
		total = data.get('total')
  
		self.num_of_processed_workers = self.num_of_processed_workers + 1
		current = self.num_of_processed_workers  
		progress_percentage = progress(current=current, total=total)
  
		if event.verbose:
			self.LOGGER.info(
				f'{progress_percentage}% ({current:,}/{total:,}) complete.'
			)

			curr_diff_time = (datetime.now() - self.updated_at).seconds
			curr_diff_time = curr_diff_time + (datetime.now() - self.updated_at).microseconds / 1e6
			self._progress_time.append(curr_diff_time)
   
			if len(self._progress_workers) == 0:
				self._progress_workers.append(current)
				self._progress_workers.append(current)
				progress_curr = current
			else:
				progress_curr = current - self._progress_workers[-1]
				self._progress_workers[-1] = progress_curr
				self._progress_workers.append(current)
    
			self._progress_workers = self._progress_workers[-6:]
			self._progress_time = self._progress_time[-5:]

			seconds_by_worker = np.array([
       			execution_time / num_of_workers 
          		for num_of_workers, execution_time in zip(self._progress_workers, self._progress_time)
            ])
   
			seconds_by_worker = seconds_by_worker.mean()

			expected_remaining_seconds = math.ceil((total - current) * seconds_by_worker)
			expected_remaining_seconds = expected_remaining_seconds + 1
   
			self.LOGGER.info(
				f'Expected remaining time: {display_time(expected_remaining_seconds)}'
			)
   
			self.updated_at = datetime.now()
   
		return lambda_response

	def get_worker_data(
		self
	) -> WorkerData:
		"""Get worker data

		Returns
		-------
		WorkerData
			The defined worker data
		"""
		return self.worker_data

	def get_result(
		self
	) -> any:
		"""Get result of all concurrent executions

		Returns
		-------
		any
			The result of all concurrent executions.
		"""
		return self.result



""" 
This module was created to load and save different types of data. 
For instance: .csv, .txt, .parquet, .gz and so on.

"""


from drtools.file_manager import (
    create_directories_of_path, search_by_name_on_directory,
    list_path_of_all_files_inside_directory
)
from typing import Union
from types import FunctionType
from pandas.core.frame import DataFrame
import pandas as pd
import os
import joblib
import pyarrow.parquet as pq


def save_df(
    dataframe: DataFrame,
    path: str,
    overwrite_if_exists: bool=False,
    **kwargs
) -> None:
    """Smart save of DataFrame.

    Parameters
    ----------
    dataframe : DataFrame
        The DataFrame to be saved.
    path : str
        Path to save dataframe
    overwrite_if_exists : bool, optional
        If True, overwrite file with same name 
        if exists. If False, return Exception 
        if exist some file on local Database with 
        same filename, by default False
    """
    if not overwrite_if_exists:
        if os.path.exists(path):
            raise Exception(f'Path {path} already exists.')
    
    create_directories_of_path(path)
    
    filename = os.path.basename(path)
    
    if '.csv' in filename:
        dataframe.to_csv(path, **kwargs)
        
    elif '.parquet' in filename:
        dataframe.to_parquet(path, **kwargs)
        
    elif '.json' in filename:
        dataframe.to_json(path, **kwargs)
        
    else:
        raise Exception(f'Extension on {os.path.basename(path)} not allow.')


def read_as_df(
    file_reference: str, 
    directory: str=None,
    **read_args,
) -> Union[DataFrame, None]:
    """Get DataFrame from name of file inside directory

    Parameters
    ----------
    file_reference : str
        Name of file that will be searched or abs path of file.
    directory : str, optional
        Directory to search for file, by default None.
    read_args : dict, optional
        Arguments that will be passed to read 
        file function, by default {}.

    Returns
    -------
    Union[DataFrame, None]
        If file is found and type is csv or parquet, returns the DataFrame, 
        else return **None**
    """
    file_path = file_reference
    if directory is not None:
        file_names = search_by_name_on_directory(
            file_reference, 
            directory
        )
        if len(file_names) == 0:
            return None
        file_path = os.path.join(directory, file_names[0])
        
    if '.csv' in file_path:
        return pd.read_csv(
            file_path,
            **read_args
        )
        
    elif '.parquet' in file_path:
        chunksize = read_args.get('chunksize', None)
        nrows = read_args.get('nrows', None)
        if chunksize is not None:
            return pq.ParquetFile(file_path)
        elif nrows is not None:
            resp = pq.ParquetFile(file_path)
            df = None
            for chunk in resp.iter_batches(batch_size=nrows):
                df = chunk.to_pandas()
                break
            usecols = read_args.get('usecols', None)
            if usecols is not None:
                df = df.loc[:, usecols]
            return df
        else:
            usecols = read_args.get('usecols', None)
            return pd.read_parquet(
                file_path,
                columns=usecols
            )
    
    elif '.json' in file_path:
        df = pd.read_json(file_path)        
        usecols = read_args.get('usecols', None)
        if usecols is not None:
            df = df.loc[:, usecols]
        return df
    
    
    elif '.joblib' in file_path:
        return joblib.load(file_path)
    
    else:
        return None
    
    
def read_dir_as_df(
    directory_path: str,
    process_chunk: FunctionType=None,
    **read_args,
) -> Union[DataFrame, None]: 
    """Load data of entire directory and load as DataFrame.

    Parameters
    ----------
    directory_path : str
        Path of directory.
    process_chunk : FunctionType
        If chunksize if passed as parameter, process each 
        chunck applying this function. This function must return 
        a DataFrame, by default None.

    Returns
    -------
    DataFrame
        The data loaded as DataFrame.
    """

    items_path = list_path_of_all_files_inside_directory(
        directory_path
    )

    df = None
    
    for item in items_path:
        resp = read_as_df(item, **read_args)
        if resp is None:
            continue
        
        temp_data = None
        chunksize = read_args.get('chunksize', None)
        nrows = read_args.get('nrows', None)
        
        if nrows is not None \
        and chunksize is not None:
            raise Exception('Parameter "chunksize" and "nrows" can not be provided at same time.')
        
        if chunksize is not None:
            if '.parquet' in item:
                for chunk in resp.iter_batches(batch_size=chunksize):
                    chunk = chunk.to_pandas()
                    if temp_data is None:
                        temp_data = process_chunk(chunk)
                    else:
                        temp_data = pd.concat([
                            temp_data,
                            process_chunk(chunk)
                        ], ignore_index=True)
            else:
                for chunk in resp:
                    if temp_data is None:
                        temp_data = process_chunk(chunk)
                    else:
                        temp_data = pd.concat([
                            temp_data,
                            process_chunk(chunk)
                        ], ignore_index=True)
                        
            usecols = read_args.get('usecols', None)
            if usecols is not None:
                temp_data = temp_data.loc[:, usecols]
                        
        elif nrows is not None:
            df_shape = 0 if df is None else df.shape[0]
            temp_data = resp[resp.index < nrows - df_shape]
            
        else:
            temp_data = resp
            
                
        if df is None: 
            if type(temp_data) == dict:
                df = temp_data
            else:
                df = temp_data.copy()
        else: 
            df = pd.concat([df, temp_data], ignore_index=True)
            
        if nrows is not None and df.shape[0] >= nrows:
            break
            
    return df


""" 
This module was created to handle time series stuff.

"""


from typing import List, Union
import drtools.utils as Utils
from pandas.core.frame import DataFrame
import pandas as pd
import math
import numpy as np
from drtools.logs import Log


OneDimension = int
TwoDimension = int
Axis0 = int
Axis1 = int


def shift(
    xs: np.ndarray, 
    periods: int, 
    fill_value: any=None,
    dimension: Union[OneDimension, TwoDimension]=1,
    axis: Union[Axis0, Axis1]=0
) -> np.ndarray:
    """Shift numpy array.

    Parameters
    ----------
    xs : np.ndarray
        The numpy array to shift.
    periods : int
        Number of periods to shift. Can be positive or negative.
    fill_value : any, optional
        The scalar value to use for newly introduced missing values, 
        by default None.

    Returns
    -------
    np.ndarray
        The shifted numpy array.
    """
    
    if dimension == 1:
        
        if periods >= 0:
            return np.concatenate((
                np.full(min(periods, len(xs)), fill_value), 
                xs[:-periods]
            ))
        
        else:
            return np.concatenate((
                xs[-periods:], 
                np.full(max(-periods, -len(xs)), fill_value)
            ))
        
    elif dimension == 2:
        
        if periods >= 0:
            
            if axis == 0:
                return np.concatenate(
                    (
                        np.full((min(periods, xs.shape[0]), xs.shape[1]), fill_value), 
                        xs[:-periods]
                    ),
                    axis=0
                )
            
            elif axis == 1:
                return np.concatenate(
                    (
                        np.full((xs.shape[0], min(periods, xs.shape[1])), fill_value), 
                        xs[:, :-periods]
                    ),
                    axis=1
                )
            
        else:
            
            if axis == 0:
                return np.concatenate(
                    (
                        xs[-periods:], 
                        np.full((min(abs(periods), xs.shape[0]), xs.shape[1]), fill_value)
                    ),
                    axis=0
                )
            
            elif axis == 1:
                return np.concatenate(
                    (
                        xs[:, -periods:], 
                        np.full((xs.shape[0], min(abs(periods), xs.shape[1])), fill_value)
                    ),
                    axis=1
                )
                
    raise Exception('Invalid "dimension" or "axis" option.')


def time_series_data_generator(
    dataframe: DataFrame,
    time_col: str,
    id_col: str,
    target_col: str,
    series_size: int,
    ignore_cols: List[str]=[],
    data_leak_cols: List[str]=[],
    immutable_cols: List[str]=[],
    light: bool=False,
    dtype: bool=True,
    chunksize: int=1,
    return_numpy: bool=False,
    LOGGER: Log=None
) -> DataFrame:
    """Generate data for time series prediction.py

    Parameters
    ----------
    dataframe : DataFrame
        DataFrame containing data that will be transformed 
        in time series.
    time_col : str
        Column that represent time of data.
    id_col : str
        Column to group data and generate time series
    target_col : str
        Predict column.
    series_size : int
        How many rows will be grouped on one row
    ignore_cols : List[str], optional
        List of columns to ignore when generating series data, 
        but the data about these columns are returned in the reponse without 
        any treatment or transformation, by default []
    data_leak_cols : List[str], optional
        Columns that are considered Data Leak, 
        by default []
    immutable_cols : List[str], optional
        Columns that not change accross time, by default []
    light : bool, optional
        If True will apply transformations on data 
        in the received DataFrame and do not will 
        create a copy of DataFrame, by default False.
    return_numpy : bool, optional
        If True will return the data as numpy array rather than 
        as DataFrame, this can safe memo usage, by default False.
    dtype : bool, optional
        If True, type cols after transformations.
        If False, will not type cols after transformations 
        , by default True.
    chunksize : int, optional
        Process by chunks, by default 1.
    LOGGER : Log, optional
        Logging object to debug execution, 
        by default None.
    
    Returns
    -------
    DataFrame
        The DataFrame after transform the data into Time Series.
    """
    
    LOGGER.debug(f'Preprocess data...')
    
    unique_ids_on_id_column = dataframe[id_col].unique()
    # num = math.ceil(len(unique_ids_on_id_column) / chunks)
    sub_ids = np.array_split(unique_ids_on_id_column, chunksize)
    sub_ids = [x for x in sub_ids if x.shape[0] > 0]
    sub_ids_len = len(sub_ids)
    
    
    if light:
        df = dataframe
    else:
        df = dataframe.copy()
    
    temp_columns = [id_col, time_col, target_col]
        
    real_features_columns = Utils.list_ops(
        df.columns,
        temp_columns + ignore_cols
    )
    
    df_dtypes = df.dtypes
    
    df.sort_values(by=[id_col, time_col], inplace=True)
    
    # df = df.sort_values(by=[id_col, time_col])
    
    real_ignore_cols = Utils.list_ops(ignore_cols, [id_col]) + [id_col]
    ignore_data_df = df.loc[:, real_ignore_cols]
    
    df = df.loc[:, [id_col, time_col] + real_features_columns + [target_col]]
    
    df_columns = df.columns
        
    mutable_columns = Utils.list_ops(
        real_features_columns + [target_col],
        immutable_cols
    )
    
    mutable_indexes = [df_columns.get_loc(col) for col in mutable_columns]
    
    columns_dtypes = {
        col: getattr(df_dtypes, col) for col in df_columns
    }
    ignore_data_columns_dtypes = {
        col: getattr(df_dtypes, col) for col in ignore_data_df.columns
    }
    
    inserted_columns = {}
    for i in range(series_size):
        inserted_columns = {
            **inserted_columns,
            **{
                f'{col}_n{i + 1}': getattr(df_dtypes, col) for col in mutable_columns
            }
        }
            
    columns_dtypes = {
        # **ignore_data_columns_dtypes,
        **columns_dtypes,
        **inserted_columns
    }
        
    final_columns = [col for col in columns_dtypes]
    
    final_values = None
    
    LOGGER.debug(f'Preprocess data... Done!')
    
    for chunck_num, ids in enumerate(sub_ids):
        
        LOGGER.debug(f'({chunck_num+1}/{sub_ids_len}) Process...')
        
        LOGGER.debug('Filtering data...')
        curr_df = df[df[id_col].isin(ids)]
        
        ignore_data = ignore_data_df[ignore_data_df[id_col].isin(ids)] \
            .drop(id_col, axis=1)
        ignore_data = ignore_data.values
        
        df_val = curr_df.values
            
        LOGGER.debug(f'Processing {curr_df.shape[0]:,} rows of data.')
        
        del curr_df
        LOGGER.debug('Filtering data... Done!')
        
    
        
        if LOGGER is not None:
            LOGGER.debug(f'Grouping by {id_col}...')
        
        df_grouped = np.split(
            df_val[:, [i for i in range(df_val.shape[1]) if i != id_col]], 
            np.unique(df_val[:, df_columns.get_loc(id_col)], return_index=True)[1][1:]
        )
        
        if LOGGER is not None:
            LOGGER.debug(f'Grouping by {id_col}... Done!')
            LOGGER.debug(f'Filling data...')
        
            
        for idx in range(len(df_grouped)):
            for i in range(series_size):
                temp_data = shift(
                    df_grouped[idx][:, mutable_indexes],
                    i + 1,
                    axis=0,
                    dimension=2
                )
                df_grouped[idx] = np.concatenate(
                    (df_grouped[idx], temp_data), 
                    axis=1
                )
                
        if LOGGER is not None:
            LOGGER.debug(f'Filling data... Done!')
            LOGGER.debug(f'Flating data...')
                
        df_val = np.array([group_data for group in df_grouped for group_data in group])
        del df_grouped
        
        if LOGGER is not None:
            LOGGER.debug(f'Flating data... Done!')
        
        
        if LOGGER is not None:
            LOGGER.debug(f'Generating final values array...')
        
        df_val = np.concatenate((ignore_data, df_val), axis=1)
        
        if final_values is None:
            final_values = df_val.copy()
        else:
            final_values = np.concatenate((final_values, df_val), axis=0)
        
        if LOGGER is not None:
            LOGGER.debug(f'Generating final values array... Done!')
            
        LOGGER.debug(f'({chunck_num+1}/{sub_ids_len}) Process... Done!')

    
    
    real_final_columns = ignore_cols + final_columns
    
    if len(data_leak_cols) > 0:
        if LOGGER is not None:
            LOGGER.debug(f'Removing data leak cols...')
            
        final_columns = Utils.list_ops(
            final_columns,
            data_leak_cols
        )
        
        dict_to_real_final_columns = {
            col: idx
            for idx, col in enumerate(real_final_columns)
        }
        
        real_final_columns = ignore_cols + final_columns
        
        final_cols_indexes = [
            dict_to_real_final_columns[col]
            for col in real_final_columns
        ]
        final_values = final_values[:, final_cols_indexes]
            
        if LOGGER is not None:
            LOGGER.debug(f'Removing data leak cols... Done!')
            
            
    if return_numpy:
        return final_columns, final_values
    else:
        LOGGER.debug(f'Generating final dataframe...')
        del df
        del ignore_data_df
        df = pd.DataFrame(
            final_values, 
            columns=real_final_columns
        )    
        del final_values
        LOGGER.debug(f'Generating final dataframe... Done!')
    
    if dtype:
        
        if LOGGER is not None:
            LOGGER.debug(f'Dtyping columns...')
        
        for idx, col in enumerate(df.columns):
            
            curr_idx = idx + 1
            LOGGER.debug(f'({curr_idx:,}/{df.columns.shape[0]:,}) Dtyping col {col}...')
            
            col_dtype = columns_dtypes.get(col, None)
            if col_dtype is None:
                LOGGER.debug(f'({curr_idx:,}/{df.columns.shape[0]:,}) Skipping Dtyping for col {col}...')
                continue
            try:
                df[col] = df[col].astype(col_dtype)
            except Exception as exc:
                
                text = f'Exception <{exc}> was generate when trying to apply ' 
                text += f'dtype {col_dtype} on column {col}' 
                
                if 'int' in col_dtype.name:
                    try:
                        df[col] = df[col].astype(pd.Int64Dtype())
                    except Exception as exc:
                        text = '1) ' + text    
                        text += f'\n2) Exception <{exc}> was generate when trying to apply ' 
                        text += f'dtype {pd.Int64Dtype()} on column {col}'              
                        LOGGER.error(text)
                else:
                    LOGGER.error(text)
                    
            LOGGER.debug(f'({curr_idx:,}/{df.columns.shape[0]:,}) Dtyping col {col}... Done!')
                    
        if LOGGER is not None:
            LOGGER.debug(f'Dtyping columns... Done!')            
        
    return df


def time_series_data_split(
    dataframe: DataFrame,
    time_col: str,
    target_col: str,
    sizes: List[float]=[0.5, 0.2, 0.3],
    light: bool=False,
    drop_time_col: bool=False,
) -> List[DataFrame]:
    """Split time series data in train, val, holdout, etc.

    Parameters
    ----------
    dataframe : DataFrame
        The dataframe representing the time series data.
    time_col : str
        Column that represent time of data.
    target_col : str
        Predict column.
    sizes : List[float], optional
        Percentage of splited datasets 
        , the sum of values must be equal 
        1, by default [0.5, 0.2, 0.3].
    light : bool, optional
        If True will apply transformations on data 
        in the received DataFrame and do not will 
        create a copy of DataFrame, by default False.
    drop_time_col : bool, optional
        If True, drop time column before return, 
        by default False.

    Returns
    -------
    List[DataFrame]
        List of X and y dataframes.
    """
    
    if light:
        df = dataframe
    else:
        df = dataframe.copy()
    
    index_split = []
    
    df.sort_values(by=time_col, inplace=True)
    # df = df.sort_values(by=time_col)
    
    # del df
    
    df.reset_index(drop=True, inplace=True)
    # df = df.reset_index(drop=True)
    
    curr_size = 0
    curr_index = 0
    for size in sizes:
        temp_index = math.ceil(df.shape[0] * (curr_size + size))
        index_split.append(
            (curr_index, temp_index)
        )
        curr_index = temp_index + 1
        curr_size = curr_size + size        
    
    if drop_time_col:
        df = df.drop(time_col, axis=1)        
    
    filters = []
    for idx_start, idx_end in index_split:
        filters.append(
            df.loc[idx_start:idx_end].index
        )
        
    # Split data
    
    resp = []
    
    for filter_ in filters:
        resp.append(
            df[df.index.isin(filter_)].drop(columns=[target_col])
        )
        resp.append(
            df[df.index.isin(filter_)][target_col]
        )
        # df = df[~df.index.isin(filter_)]
        
    del df
    
    return resp


""" 
This module was created to handle
default authentication.

"""


from selenium.webdriver.common.by import By
from drtools.scrapping.interaction import click
from selenium.webdriver.chrome.webdriver import WebDriver


def auto_login(
    driver: WebDriver, 
    username: str, 
    password: str
) -> WebDriver:
    """Perform automatic login

    Parameters
    ----------
    driver : WebDriver
        Selenium WebDriver
    username : str
        The username that will be placed on login form
    password : str
        The password that will be placed on password input

    Returns
    -------
    WebDriver
        Selenium WebDriver
    """
    # Find a password input field and enter the specified password string
    password_xpath = "//input[@type='password']"
    # password_input = wait_for_element(driver, EC.presence_of_element_located((By.XPATH, password_xpath)))
    password_input = driver.find_element(by=By.XPATH, value=password_xpath)
    # driver.execute_script(f"arguments[0].value = \"{password}\";", password_input)
    password_input.send_keys(password)

    # Find a visible input field preceding out password field and enter the specified username
    # username_abs_xpath = f"{password_xpath}/preceding::input[not(@type='hidden')][1]"
    # username_input = wait_for_element(driver, EC.presence_of_element_located((By.XPATH, username_abs_xpath)))
    username_xpath = "preceding::input[not(@type='hidden')][1]"
    username_input = password_input.find_element(by=By.XPATH, value=username_xpath)
    # driver.execute_script(f"arguments[0].value = \"{username}\";", username_input)
    username_input.send_keys(username)

    # Find the form element enclosing our password field
    form_element = password_input.find_element(by=By.XPATH, value=".//ancestor::form")

    # Find the form's submit element and click it
    submit_button = form_element.find_element(by=By.XPATH, value=".//*[@type='submit']")
    click(submit_button, driver=driver)

    return driver


""" 
This module was created to archive general functions, methods, 
classes and other stuff.

"""


from typing import Dict, List, Union, Optional, TypedDict
import numpy as np
from pandas import DataFrame
from drtools.utils import list_ops
from drtools.logs import Log


### Comparison Operators
EqualsTo = "$eq"
GreaterThan = "$gt"
GreaterThanOrEqual = "$gte"
In = "$in"
LessThan = "$lt"
LessThanOrEqual = "$lte"
NotEqual = "$ne"
NotIn = "$nin"

class ComparisonQueryOperators(TypedDict):
    EqualsTo: Optional[Union[float, str]]
    GreaterThan: Optional[float]
    GreaterThanOrEqual: Optional[float]
    In: Optional[List[Union[float, str]]]
    LessThan: Optional[float]
    LessThanOrEqual: Optional[float]
    NotEqual: Optional[Union[float, str]]
    NotIn: Optional[List[Union[float, str]]]


FieldName = str
FieldQuery = Dict[FieldName, ComparisonQueryOperators]


### Logical Operators
And = "$and"
Or = "$or"

class LogicalQueryOperators(TypedDict):
    And: Optional[Union[List, List[FieldQuery]]]
    Or: Optional[Union[List, List[FieldQuery]]]


Query = Union[LogicalQueryOperators, ComparisonQueryOperators]


NumpyDataType = 'numpy'
DataFrameDataType = 'dataframe'
Data = Union[DataFrame, np.array]


class FindOnData:
    """Query for data on Numpy matrix or Pandas DataFrame.
    """
    
    def __init__(
        self,
        data: Data,
        query: Query,
        data_type: Union[DataFrameDataType, NumpyDataType]=DataFrameDataType,
        data_columns: List[str]=None,
        LOGGER=None
    ) -> None:        
        assert data_type in [DataFrameDataType, NumpyDataType], \
            f'Invalid data_type: {data_type}'
        if data_type == NumpyDataType:
            assert data_columns is not None and len(data_columns) > 0, \
                f'When data_type equals to {data_type}, "data_columns" must be provided.'
        self.Query = query
        self.Data = data
        self.DataType = data_type
        self.DataColumns = data_columns
        if self.DataColumns is not None:
            self.DataColNameToIdx = {
                col_name: idx
                for idx, col_name in enumerate(self.DataColumns)
            }
        if LOGGER is None:
            self.LOGGER = Log()
        else:
            self.LOGGER = LOGGER
    
    def _is_valid_logical_syntax(self, query) -> bool:
        try:
            if And in query:
                if Or in query:
                    return False
                if isinstance(query[And], list):
                    return True
                else:
                    return False
            elif Or in query:
                if isinstance(query[Or], list):
                    return True
                else:
                    return False
            else:
                keys_list = list(query.keys())
                if len(keys_list) == 0:
                    return True
                else:
                    return False
        except:
            return False
    
    def _is_valid_comparison_syntax(self, query) -> bool:
        if not isinstance(query, dict):
            return False
        
        try:
            for k, v in query.items():
                keys_list = list(v.keys())
                not_expected_keys = list_ops(
                    keys_list,
                    [
                        EqualsTo, GreaterThan, GreaterThanOrEqual, 
                        In, LessThan, LessThanOrEqual, NotEqual, NotIn
                    ]
                )
                if len(not_expected_keys) > 0:
                    return False
            return True
        except:
            return False
    
    
    ##########################################
    ### Comparison Operators
    ##########################################
    def _is_eq_op(self, op_name):
        return op_name == "$eq"
    
    def _is_gt_op(self, op_name):
        return op_name == "$gt"
    
    def _is_gte_op(self, op_name):
        return op_name == "$gte"
    
    def _is_in_op(self, op_name):
        return op_name == "$in"
    
    def _is_lt_op(self, op_name):
        return op_name == "$lt"
    
    def _is_lte_op(self, op_name):
        return op_name == "$lte"
    
    def _is_ne_op(self, op_name):
        return op_name == "$ne"
    
    def _is_nin_op(self, op_name):
        return op_name == "$nin"
    
    def _is_isna_op(self, op_name):
        return op_name == "$isna"
    
    def _perform_eq_op(self, column, value):
        if self.DataType == DataFrameDataType:
            return self.Data[column] == value
        elif self.DataType == NumpyDataType:
            return self.Data[:, column] == value
    
    def _perform_gt_op(self, column, value):
        if self.DataType == DataFrameDataType:
            return self.Data[column] > value
        elif self.DataType == NumpyDataType:
            return self.Data[:, column] > value
    
    def _perform_gte_op(self, column, value):
        if self.DataType == DataFrameDataType:
            return self.Data[column] >= value
        elif self.DataType == NumpyDataType:
            return self.Data[:, column] >= value
    
    def _perform_in_op(self, column, value: List):
        if self.DataType == DataFrameDataType:
            return self.Data[column].isin(value)
        elif self.DataType == NumpyDataType:
            raise Exception("Operation $in must for NumpyDataType.")
    
    def _perform_lt_op(self, column, value):
        if self.DataType == DataFrameDataType:
            return self.Data[column] < value
        elif self.DataType == NumpyDataType:
            return self.Data[:, column] < value
    
    def _perform_lte_op(self, column, value):
        if self.DataType == DataFrameDataType:
            return self.Data[column] <= value
        elif self.DataType == NumpyDataType:
            return self.Data[:, column] <= value
    
    def _perform_ne_op(self, column, value):
        if self.DataType == DataFrameDataType:
            return self.Data[column] != value
        elif self.DataType == NumpyDataType:
            return self.Data[:, column] != value
    
    def _perform_nin_op(self, column, value):
        if self.DataType == DataFrameDataType:
            return ~self.Data[column].isin(value)
        elif self.DataType == NumpyDataType:
            raise Exception("Operation $nin must for NumpyDataType.")
        
    def _perform_isna_op(self, column, value):
        if self.DataType == DataFrameDataType:
            if value is True:
                return self.Data[column].isna()
            else:
                return self.Data[column].notna()
        elif self.DataType == NumpyDataType:
            raise Exception("Operation $isna must for NumpyDataType.")
    
    def _perform_comparison_operation(self, data, query) -> DataFrame:
        final_conditions_response = None
        resp = None
        for col, single_comparison_query in query.items():
            for operation, val in single_comparison_query.items():
                if self._is_eq_op(op_name=operation):
                    resp = self._perform_eq_op(column=col, value=val)
                elif self._is_gt_op(op_name=operation):
                    resp = self._perform_gt_op(column=col, value=val)
                elif self._is_gte_op(op_name=operation):
                    resp = self._perform_gte_op(column=col, value=val)
                elif self._is_in_op(op_name=operation):
                    resp = self._perform_in_op(column=col, value=val)
                elif self._is_lt_op(op_name=operation):
                    resp = self._perform_lt_op(column=col, value=val)
                elif self._is_lte_op(op_name=operation):
                    resp = self._perform_lte_op(column=col, value=val)
                elif self._is_ne_op(op_name=operation):
                    resp = self._perform_ne_op(column=col, value=val)
                elif self._is_nin_op(op_name=operation):
                    resp = self._perform_nin_op(column=col, value=val)
                elif self._is_isna_op(op_name=operation):
                    resp = self._perform_isna_op(column=col, value=val)
                else:
                    raise Exception(f"Invalid comparison operator: {single_comparison_query}")
                if final_conditions_response is None:
                    final_conditions_response = resp
                else:
                    final_conditions_response = self._perform_and_operation(
                        conditional1=final_conditions_response,
                        conditional2=resp
                    )
        return final_conditions_response
    
    
    ##########################################
    ### Logical Operators
    ##########################################
    def _is_and_syntax(self, query):
        if And in query:
            return True
        return False
    
    def _is_or_syntax(self, query):
        if Or in query:
            return True
        return False
        
    def _perform_and_operation(self, conditional1, conditional2):
        return conditional1 & conditional2
        
    def _perform_or_operation(self, conditional1, conditional2):
        return conditional1 | conditional2
    
    def _perform_logical_operation(self, conditional1: List, conditional2: List, query) -> DataFrame:
        resp = None
        if self._is_and_syntax(query=query):
            resp = self._perform_and_operation(conditional1=conditional1, conditional2=conditional2)
        elif self._is_or_syntax(query=query):
            resp = self._perform_or_operation(conditional1=conditional1, conditional2=conditional2)
        return resp
    
    ##########################################
    ### Find
    ##########################################
     
    def _find_on_numpy(self):
        def _recursive_query(query, depth: int=0):
            if self._is_valid_logical_syntax(query=query):
                final_query_conditional_response = None
                for k, v in query.items():
                    for expression in v:
                        conditional_response = _recursive_query(query=expression, depth=depth+1)
                        if final_query_conditional_response is None:
                            final_query_conditional_response = conditional_response
                        else:
                            final_query_conditional_response = self._perform_logical_operation(
                                conditional1=final_query_conditional_response,
                                conditional2=conditional_response,
                                query=query
                            )
                return final_query_conditional_response
            elif self._is_valid_comparison_syntax(query=query):
                real_query = {}
                for k, v in query.items():
                    real_query[int(self.DataColNameToIdx[k])] = v
                resp = self._perform_comparison_operation(
                    data=self.Data, 
                    query=real_query
                )
                return resp
        query_conditional = _recursive_query(query=self.Query, depth=0)
        return self.Data[query_conditional, :]
    
    def _find_on_dataframe(self):
        def _recursive_query(query, depth: int=0):
            if self._is_valid_logical_syntax(query=query):
                final_query_conditional_response = None
                for k, v in query.items():
                    for expression in v:
                        conditional_response = _recursive_query(query=expression, depth=depth+1)
                        if final_query_conditional_response is None:
                            final_query_conditional_response = conditional_response
                        else:
                            final_query_conditional_response = self._perform_logical_operation(
                                conditional1=final_query_conditional_response,
                                conditional2=conditional_response,
                                query=query
                            )
                return final_query_conditional_response
            elif self._is_valid_comparison_syntax(query=query):
                self.LOGGER.info(f'Comparison Query: {query}')
                resp = self._perform_comparison_operation(
                    data=self.Data, 
                    query=query
                )
                self.LOGGER.info(f'Valid rows: {resp.sum()} from {self.Data.shape[0]}')
                return resp
        query_conditional = _recursive_query(query=self.Query, depth=0)
        return self.Data[query_conditional]
    
    def run_query(self) -> Data:
        if self.DataType == NumpyDataType:
            return self._find_on_numpy()
        elif self.DataType == DataFrameDataType:
            return self._find_on_dataframe()


""" 
This module was created to manage logs of executions
of any .py or .ipynb file

"""


import sys
import logging
from types import FunctionType
from typing import Any, Union, TypedDict
from drtools.file_manager import (
    split_path, create_directories_of_path
)
from functools import wraps
import logging
from logging.handlers import RotatingFileHandler
from inspect import getframeinfo, stack


class CallerFilter(logging.Filter):
    """ This class adds some context to the log record instance """
    file = ''
    line_n = ''

    def filter(self, record):
        record.file = self.file
        record.line_n = self.line_n
        return True


def caller_reader(f):
    """This wrapper updates the context with the callor infos"""
    def wrapper(self, *args):
        caller = getframeinfo(stack()[1][0])
        last_name = split_path(
            caller.filename
        )[-1]
        file = caller.filename \
            if self.full_file_path_log \
            else last_name
        line_n = caller.lineno
        self._filter.file = f'{file}:{line_n}'
        return f(self, *args)
    wrapper.__doc__ = f.__doc__
    return wrapper


class FormatterOptions(TypedDict):
    IncludeThreadName: bool
    IncludeFileName: bool
    IncludeDate: bool
    IncludeLoggerName: bool
    IncludeLevelName: bool


class Log:
    """Handle logging
    
    Note
    -----
    You can use the max_bytes and backup_count values to allow 
    the file to rollover at a predetermined size. When the 
    size is about to be exceeded, the file is closed and 
    a new file is silently opened for output. Rollover occurs 
    whenever the current log file is nearly max_bytes in 
    length; but if either of max_bytes or backup_count is 
    zero, rollover never occurs, so you generally want 
    to set backup_count to at least 1, and have a non-zero 
    max_bytes. When backup_count is non-zero, the system 
    will save old log files by appending the 
    extensions '.1', '.2' etc., to the filename. For example, with 
    a backup_count of 5 and a base file name of app.log, you 
    would get app.log, app.log.1, app.log.2, up to app.log.5. The 
    file being written to is always app.log. When this file is 
    filled, it is closed and renamed to app.log.1, and if files 
    app.log.1, app.log.2, etc. exist, then they are renamed to 
    app.log.2, app.log.3 etc. respectively.
    
    Parameters
    ----------
    path : str
        Path to save logs
    max_bytes : int, optional
        Max bytes which one log file
        will be at maximum, by default 2*1024*1024
    backup_count : int, optional
        Number of backup logs that will be
        alive at maximum, by default 5
    name : str, optional
        Logger name, by default 'Open-Capture'
    default_start : bool, optional
        Log the initialization, by default True
    full_file_path_log : bool, optional
        If True, log file path will be complete
        If False, only will be displayed the name
        of the file, by default False
    """

    def __init__(
        self,
        path: Union[str, None]=None,
        max_bytes: int=2 * 1024 * 1024,
        backup_count: int=10,
        name: str='Main',
        default_start: bool=True,
        full_file_path_log: bool=False,
        log_level: str='DEBUG',
        formatter_options: FormatterOptions=FormatterOptions({
            'IncludeThreadName': True,
            'IncludeFileName': True,
            'IncludeDate': True,
            'IncludeLoggerName': True,
            'IncludeLevelName': True,
        }),
        # deprecated 
        log_as_print: bool=False
    ) -> None:
        self.path = path
        self.max_bytes = max_bytes
        self.backup_count = backup_count
        self.name = name
        self.default_start = default_start
        self.full_file_path_log = full_file_path_log
        self.log_level = log_level
        self.formatter_options = formatter_options
        self.log_level = {'DEBUG': 10, 'INFO': 20, 'WARNING': 30, 'ERROR': 40, 'CRITICAL': 50}[self.log_level]        
        self.original_verbosity = self.log_level
        self._construct_logger()
        if default_start:
            self.info('!*************** START ***************!')

    def _construct_logger(
        self,
    ):
        self.LOGGER = logging.getLogger(self.name)
        if self.LOGGER.hasHandlers():
            self.LOGGER.handlers.clear() # Clear the handlers to avoid double logs        
        if self.path is not None:
            create_directories_of_path(self.path)
            log_handler = RotatingFileHandler(
                self.path, 
                mode='a', 
                maxBytes=self.max_bytes,
                backupCount=self.backup_count, 
                encoding=None, 
                delay=0
            )
        else:
            log_handler = logging.StreamHandler(sys.stdout)
        formatter_text = ''
        accept_conditions = [True]
        if self.formatter_options.get('IncludeThreadName', None) in accept_conditions:
            formatter_text = '[%(threadName)-14s] '
        if self.formatter_options.get('IncludeFileName', None) in accept_conditions:
            formatter_text = formatter_text + '[%(file)-20s] '
        if self.formatter_options.get('IncludeDate', None) in accept_conditions:
            formatter_text = formatter_text + '[%(asctime)s.%(msecs)03d] '
        if self.formatter_options.get('IncludeLoggerName', None) in accept_conditions:
            formatter_text = formatter_text + '[%(name)-12s] '
        if self.formatter_options.get('IncludeLevelName', None) in accept_conditions:
            formatter_text = formatter_text + '[%(levelname)8s] '        
        formatter = logging.Formatter(formatter_text + '%(message)s', datefmt='%d-%m-%Y %H:%M:%S')
        log_handler.setFormatter(formatter)
        self.LOGGER.addHandler(log_handler)
        # Here we add the Filter, think of it as a context
        self._filter = CallerFilter()
        self.LOGGER.addFilter(self._filter)
        self.LOGGER.setLevel(self.log_level)

    def set_verbosity(self, verbosity: bool=True) -> None:
        """Set verbosity of logs.

        Parameters
        ----------
        verbosity : bool, optional
            If True, log all levels, 
            If False, log nothing, by default True
        """
        if verbosity:
            self.log_level = 10
        else:
            self.log_level = 999
        self.LOGGER.setLevel(self.log_level)

    def reset_verbosity(self) -> None:
        """Turn verbosity as initial state.
        """
        self.log_level = self.original_verbosity
        self.LOGGER.setLevel(self.log_level)
    
    @caller_reader
    def debug(self, msg: any) -> None:
        """Log in DEBUG level

        Parameters
        ----------
        msg : any
            The message that will be logged
        """
        self.LOGGER.debug(msg)

    @caller_reader
    def info(self, msg: any) -> None:
        """Log in INFO level

        Parameters
        ----------
        msg : any
            The message that will be logged
        """
        self.LOGGER.info(msg)
        
    @caller_reader
    def warning(self, msg: any) -> None:
        """Log in WARNING level

        Parameters
        ----------
        msg : any
            The message that will be logged
        """
        self.LOGGER.warning(msg)

    @caller_reader
    def error(self, msg: any) -> None:
        """Log in ERROR level

        Parameters
        ----------
        msg : any
            The message that will be logged
        """
        self.LOGGER.error(msg)
        
    @caller_reader
    def critical(self, msg: any) -> None:
        """Log in CRITICAL level

        Parameters
        ----------
        msg : any
            The message that will be logged
        """
        self.LOGGER.critical(msg)
    

def function_name_start_and_end(
    func: FunctionType,
    logger: Log
) -> FunctionType:
    """Log name of function.
    
    Logs the name of function on start and end of execution.
    Logs error too.

    Parameters
    ----------
    func : FunctionType
        Function that will be executed
    logger : Logger, optional
        Specific logger, by default logging

    Returns
    -------
    FunctionType
        The wrapper function.
    """
    @wraps(func)
    def wrapper(*args, **kwargs) -> Union[Any, None]:
        logger.debug(f'FunctionExecution : Start : {func.__name__}()')
        response = None
        try:
            response = func(*args, **kwargs)
        except Exception as exc:
            logger.error(exc)
        logger.debug(f'FunctionExecution : End : {func.__name__}')
        return response
    return wrapper